from typing import Any

from databricks.ml_features_common.entities._proto_enum_entity import _ProtoEnumEntity
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    FeatureDefinitionType as ProtoFeatureDefinitionType,
)


class FeatureDefinitionType(_ProtoEnumEntity):
    """Feature definition types supported when serving features."""

    # Should be handled the same way as FEATURE_TABLE
    FEATURE_DEFINITION_TYPE_UNSPECIFIED = ProtoFeatureDefinitionType.Value(
        "FEATURE_DEFINITION_TYPE_UNSPECIFIED"
    )
    # Lookup Client will depend on both serialized FeatureTablesForServing and the model's FeatureSpec.
    FEATURE_TABLE = ProtoFeatureDefinitionType.Value("FEATURE_TABLE")
    # Lookup Client will only depend on the serialized FeatureTablesForServing.
    DECLARATIVE_FEATURES = ProtoFeatureDefinitionType.Value("DECLARATIVE_FEATURES")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoFeatureDefinitionType

    @classmethod
    def from_proto_value(cls, proto_value: int) -> int:
        cls.init()
        if proto_value not in cls._ENUM_TO_STRING:
            raise ValueError(
                f"Unsupported FeatureDefinitionType proto value: {proto_value}. "
                f"Valid values: {list(cls._ENUM_TO_STRING.keys())}"
            )
        return proto_value



import collections
from typing import Optional, Tuple

import sqlalchemy

from databricks.feature_store.lookup_engine.lookup_sql_engine import LookupSqlEngine
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)


class LookupSqlServerEngine(LookupSqlEngine):
    def __init__(
        self, online_feature_table: OnlineFeatureTable, ro_user: str, ro_password: str
    ):
        super().__init__(online_feature_table, ro_user, ro_password)

    @property
    def engine_url(self):
        return f"mssql+pyodbc://{self.user}:{self.password}@{self.host}:{self.port}/{self.database_name}?driver=ODBC+Driver+17+for+SQL+Server"

    @classmethod
    def _sql_safe_name(cls, name):
        # MSSQL requires [xxx] format to safely handle identifiers that contain special characters or are reserved words.
        return f"[{name}]"

    def _database_contains_feature_table(self, sql_connection):
        query = sqlalchemy.sql.text(
            f"SELECT {self.TABLE_NAME} FROM {self.INFORMATION_SCHEMA}.{self.TABLES} "
            f"WHERE {self.TABLE_CATALOG}='{self.database_name}' AND {self.TABLE_NAME} IN ('{self.table_name}')"
        )
        results = sql_connection.execute(query)
        table = results.fetchall()
        return len(table) > 0

    def _database_contains_primary_keys(self, sql_connection):
        query = sqlalchemy.sql.text(
            f"SELECT col.{self.COLUMN_NAME} FROM {self.INFORMATION_SCHEMA}.{self.TABLE_CONSTRAINTS} tab, {self.INFORMATION_SCHEMA}.{self.CONSTRAINT_COLUMN_USAGE} col "
            f"WHERE tab.{self.TABLE_CATALOG}='{self.database_name}' AND col.{self.TABLE_CATALOG}='{self.database_name}' AND "
            f"col.{self.CONSTRAINT_NAME}=tab.{self.CONSTRAINT_NAME} AND col.{self.TABLE_NAME}=tab.{self.TABLE_NAME} AND "
            f"{self.CONSTRAINT_TYPE}='PRIMARY KEY' AND col.{self.TABLE_NAME}='{self.table_name}'"
        )
        results = sql_connection.execute(query)
        primary_keys = [r[0] for r in results.fetchall()]
        return collections.Counter(primary_keys) == collections.Counter(
            [primary_key.name for primary_key in self.primary_keys]
        )

    # Override
    def _get_database_and_table_name(
        self, online_table_name
    ) -> Tuple[str, Optional[str], str]:
        name_components = online_table_name.split(".")
        if len(name_components) != 2:
            raise ValueError(
                f"Online table name {online_table_name} is misformatted and must be in 2L format for SqlServer stores"
            )
        return (name_components[0], None, name_components[1])



from typing import Callable


class LatencyDebugger:
    def __init__(self, current_time_seconds_fn: Callable[[], float]):
        # list of tuples (tag, elapsed_time) in milliseconds
        self.enabled = False
        self.current_time_fn = current_time_seconds_fn
        self.elapsed_times = None

    def init(self):
        self.enabled = True
        self.elapsed_times = [("start", self._current_time_ms())]

    def add_latency(self, tag: str):
        if self.enabled:
            self.elapsed_times.append((tag, self._current_time_ms()))

    def debug_string(self) -> str:
        if not self.enabled:
            return ""
        start = self.elapsed_times[0][1]
        latencies = [(tag, t - start) for (tag, t) in self.elapsed_times]
        return f"Latency Debugger: {latencies}"

    def _current_time_ms(self):
        return int(self.current_time_fn() * 1000)



"""
Abstract metrics recording interface for feature store lookup client.

This module defines the interface but has ZERO dependencies on any metrics
backend (Prometheus). Implementations are provided by the runtime environment.

TODO: Ideally, this should be a library shared with model serving so there is a 
clear contract. Right now, scoring server just assumes the APIs are provided.
"""

from abc import ABC, abstractmethod
from typing import Dict, Optional


class MetricsRecorder(ABC):
    """
    Abstract interface for metrics recording.

    The lookup client emits metrics by calling methods on this interface.
    The runtime environment (model server, notebook, Lambda) provides
    the concrete implementation.
    """

    @abstractmethod
    def record_counter(
        self, name: str, value: float, labels: Optional[Dict[str, str]] = None
    ):
        """
        Record a counter metric (monotonically increasing).

        :param name: Metric name (e.g., "databricks.feature_store.lookup.count")
        :param value: Value to add to counter
        :param labels: Label dimensions (e.g., {"table": "users", "status": "success"})
        """
        pass

    @abstractmethod
    def record_histogram(
        self, name: str, value: float, labels: Optional[Dict[str, str]] = None
    ):
        """
        Record a histogram metric (for latency, size, etc.).

        Histograms buckets are defined by model server (go/model-serving-metric-buckets).
        The latency metrics are estimated from the buckets with acceptable errors. See
        https://prometheus.io/docs/practices/histograms/#quantiles for how quantiles can
        be estimated.

        :param name: Metric name (e.g., "feature_lookup_client_feature_store_latency_ms")
        :param value: Observed value (e.g., latency in milliseconds)
        :param labels: Label dimensions (e.g., {"table": "users"})
        """
        pass


class _NoOpRecorder(MetricsRecorder):
    """
    No-op implementation used when no recorder is registered.

    This ensures zero overhead when metrics are disabled or in
    environments that don't support metrics (notebooks, local dev).
    """

    def record_counter(
        self, name: str, value: float, labels: Optional[Dict[str, str]] = None
    ):
        pass

    def record_histogram(
        self, name: str, value: float, labels: Optional[Dict[str, str]] = None
    ):
        pass



import logging
import os
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd

from databricks.feature_store.lookup_engine import (
    AwsAccessKey,
    LookupCosmosDbEngine,
    LookupDynamoDbEngine,
    LookupLakebaseEngine,
    LookupLegacyBrickstoreHttpEngine,
    LookupMySqlEngine,
    LookupSqlEngine,
    LookupSqlServerEngine,
)
from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.lakebase_constants import (
    BRICKSTORE_OAUTH_TOKEN_FILE_PATH,
)
from databricks.feature_store.utils.lakebase_utils import (
    get_lakebase_connection_pool_size,
)
from databricks.feature_store.utils.logging_utils import get_logger
from databricks.feature_store.utils.lookup_client_envvars import DISABLE_CONNECTION_POOL
from databricks.feature_store.utils.metrics_utils import LookupClientMetrics
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)
from databricks.ml_features_common.entities.query_mode import QueryMode
from databricks.ml_features_common.entities.store_type import StoreType

_logger = get_logger(__name__, log_level=logging.INFO)


# The provisioner of this model is expected to set the following environment variable for each
# feature table if the feature store is SQL based:
#  (1) <online_store_for_serving.read_secret_prefix>_USER
#  (2) <online_store_for_serving.read_secret_prefix>_PASSWORD
USER_SUFFIX = "_USER"
PASSWORD_SUFFIX = "_PASSWORD"
# Prefix for Lakebase user and password from environment variable
LAKEBASE_PREFIX = "LAKEBASE"
# For DynamoDB the following variables should be set:
#  (1) <online_store_for_serving.read_secret_prefix>_ACCESS_KEY_ID
#  (2) <online_store_for_serving.read_secret_prefix>_SECRET_ACCESS_KEY
ACCESS_KEY_ID_SUFFIX = "_ACCESS_KEY_ID"
SECRET_ACCESS_KEY_SUFFIX = "_SECRET_ACCESS_KEY"
# For Cosmos DB the following variable should be set:
#  (1) <online_store_for_serving.read_secret_prefix>_AUTHORIZATION_KEY
AUTHORIZATION_KEY_SUFFIX = "_AUTHORIZATION_KEY"
LookupKeyType = Tuple[str, ...]


def generate_lookup_sql_engine(
    online_feature_table: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
    creds: Tuple[str, str],
    total_table_count: Optional[int],
) -> LookupSqlEngine:
    if isinstance(online_feature_table, List):
        raise Exception(f"Internal error: Batch lookup is unsupported in SQL Engine.")

    ro_user, ro_password = creds
    if (
        online_feature_table.online_store.store_type == StoreType.BRICKSTORE
        or online_feature_table.online_store.store_type
        == StoreType.DATABRICKS_ONLINE_STORE
    ):
        lakebase_connection_pool_size = get_lakebase_connection_pool_size(
            total_table_count
        )
        return LookupLakebaseEngine(
            online_feature_table,
            ro_user,
            ro_password,
            lakebase_connection_pool_size,
        )
    elif online_feature_table.online_store.store_type == StoreType.SQL_SERVER:
        return LookupSqlServerEngine(online_feature_table, ro_user, ro_password)
    return LookupMySqlEngine(online_feature_table, ro_user, ro_password)


def generate_lookup_dynamodb_engine(
    online_feature_table: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
    creds: Optional[Tuple[str, str]],
) -> LookupDynamoDbEngine:
    if creds:
        access_key_id, secret_access_key = creds

        return LookupDynamoDbEngine(
            online_feature_table,
            access_key=AwsAccessKey(access_key_id, secret_access_key),
        )
    else:
        return LookupDynamoDbEngine(online_feature_table, access_key=None)


def generate_lookup_cosmosdb_engine(
    online_feature_table: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
    creds: str,
) -> LookupCosmosDbEngine:
    if isinstance(online_feature_table, List):
        raise Exception(f"Internal error: Batch lookup is unsupported in Cosmos DB.")
    return LookupCosmosDbEngine(online_feature_table, authorization_key=creds)


def generate_lookup_legacy_brickstore_http_engine(
    online_feature_table: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
) -> LookupLegacyBrickstoreHttpEngine:
    disable_connection_pool = os.environ.get(DISABLE_CONNECTION_POOL, "false") == "true"
    return LookupLegacyBrickstoreHttpEngine(
        online_feature_table, disable_connection_pool
    )


def load_credentials_from_env(online_ft: OnlineFeatureTable):
    def get_env_var(env_var_suffix: str) -> str:
        env_var = ""
        if (
            online_ft.online_store.store_type == StoreType.BRICKSTORE
            or online_ft.online_store.store_type == StoreType.DATABRICKS_ONLINE_STORE
        ):
            env_var = LAKEBASE_PREFIX + env_var_suffix
        else:
            env_var = online_ft.online_store.read_secret_prefix + env_var_suffix
            if env_var not in os.environ:
                raise Exception(
                    f"Internal error: {env_var} not found for feature table {online_ft.feature_table_name}."
                )
        return os.getenv(env_var)

    if online_ft.online_store.store_type == StoreType.DYNAMODB:
        try:
            return (
                get_env_var(ACCESS_KEY_ID_SUFFIX),
                get_env_var(SECRET_ACCESS_KEY_SUFFIX),
            )
        except:
            # It is assumed that if the secrets are not configured, the lookup will be authorized
            # using instance profile. This is validated in serving-scheduler on endpoint creation
            # see EndpointDeploymentProvisioner.verifyAndMaybeUpdateFsMetadataForInstanceProfile
            # TODO (ML-33307): add env var check whether instance profile is configured
            print(
                "Secret credentials not configured, attempting to use cluster instance profile."
            )
            return None
    elif online_ft.online_store.store_type == StoreType.COSMOSDB:
        return get_env_var(AUTHORIZATION_KEY_SUFFIX)
    return (
        get_env_var(USER_SUFFIX),
        get_env_var(PASSWORD_SUFFIX),
    )


# TODO(ML-26146): refactor this method and above helpers to appropriate seperate util files
def tables_share_dynamodb_access_keys(
    online_feature_tables: List[OnlineFeatureTable],
):
    # check region and credentials to see if all feature tables are using same DynamoDB accesss keys
    region, creds = None, None
    for ft in online_feature_tables:
        if ft.online_store.store_type != StoreType.DYNAMODB:
            return False
        elif region == None:
            region = ft.online_store.extra_configs.region
            creds = load_credentials_from_env(ft)
        elif (
            region != ft.online_store.extra_configs.region
            or load_credentials_from_env(ft) != creds
        ):
            return False
    return True


def can_use_brickstore_http_gateway(
    online_feature_tables: Union[OnlineFeatureTable, List[OnlineFeatureTable]]
):
    if isinstance(online_feature_tables, OnlineFeatureTable):
        online_feature_tables = [online_feature_tables]
    return (
        all(
            [
                oft.online_store.store_type == StoreType.BRICKSTORE
                or oft.online_store.store_type == StoreType.DATABRICKS_ONLINE_STORE
                for oft in online_feature_tables
            ]
        )
        and len(online_feature_tables) > 0
        and online_feature_tables[0].online_store.extra_configs.table_serving_url != ""
        and os.path.exists(BRICKSTORE_OAUTH_TOKEN_FILE_PATH)
    )


def is_primary_key_lookup(
    online_feature_tables: List[OnlineFeatureTable],
):
    for oft in online_feature_tables:
        mode = oft.online_store.query_mode
        if mode == QueryMode.RANGE_QUERY:
            return False
        elif mode == QueryMode.PRIMARY_KEY_LOOKUP:
            continue
        else:
            raise ValueError(f"Unsupported query mode: {mode}")
    return True


class OnlineLookupClient:
    def __init__(
        self,
        online_feature_table: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
        total_table_count: Optional[int],
        eligible_for_pgsql: bool = False,
    ):
        """
        :param online_feature_table: The online feature table to lookup.
        :param total_table_count: Total number of tables queried by this endpoint required for
            Lakebase PgSql connections because it creates connection pool based on the number
            of tables.
        :param eligible_for_pgsql: Whether the online feature table is eligible for Lakebase PgSql.
        """
        self.lookup_engine = self._generate_lookup_engine(
            online_feature_table, eligible_for_pgsql, total_table_count
        )

    @classmethod
    def _generate_lookup_engine(
        cls,
        online_feature_table: Union[OnlineFeatureTable, List[OnlineFeatureTable]],
        eligible_for_pgsql: bool,
        total_table_count: Optional[int],
    ):
        first_online_feature_table = None
        if isinstance(online_feature_table, List):
            if len(online_feature_table) == 0:
                raise Exception(
                    f"Internal Error: No feature table passed while creating lookup engine."
                )
            first_online_feature_table = online_feature_table[0]
        else:
            first_online_feature_table = online_feature_table

        creds = load_credentials_from_env(first_online_feature_table)
        if first_online_feature_table.online_store.store_type == StoreType.DYNAMODB:
            return generate_lookup_dynamodb_engine(online_feature_table, creds)
        elif first_online_feature_table.online_store.store_type == StoreType.COSMOSDB:
            return generate_lookup_cosmosdb_engine(online_feature_table, creds)
        elif (
            first_online_feature_table.online_store.store_type
            in {StoreType.BRICKSTORE, StoreType.DATABRICKS_ONLINE_STORE}
            and can_use_brickstore_http_gateway(online_feature_table)
            and not eligible_for_pgsql
        ):
            return generate_lookup_legacy_brickstore_http_engine(online_feature_table)
        return generate_lookup_sql_engine(
            online_feature_table, creds, total_table_count
        )

    def batch_lookup_features(
        self,
        lookup_list_dict: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_names_dict: Dict[str, Dict[LookupKeyType, List[str]]],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        """
        :param lookup_list_dict: dictionary from online feature table name to a dictionary from
            lookup key to primary key list.
        :param feature_names_dict:  dictionary from online feature table name to feature names.
        :return a Dictionary from online feature table name to feature values.
        """
        return self.lookup_engine.batch_lookup_features(
            lookup_list_dict, feature_names_dict, metrics=metrics
        )

    def lookup_features(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        """Uses a Python database connection to lookup features in feature_names using
        the lookup keys and values in lookup_list. The online store database and table name are
        obtained from the OnlineFeatureTable passed to the constructor.

        The resulting DataFrame has the same number of rows as lookup_list. In the case that a
        lookup key cannot be found, a row of NaNs will be returned in the resulting DataFrame.

        Throws an exception if the table, lookup keys, or feature columns do not exist in the
        online store.

        :param lookup_list: Should contain one column for each primary key of the online feature
        table, and one row for each entity to look up.
        :param feature_names: A list of feature names to look up.
        :return: Pandas DataFrame containing feature keys and values fetched from the online store.
        """
        features = self.lookup_engine.lookup_features(
            lookup_list, feature_names, metrics=metrics
        )
        return features

    def lookup_feature_dicts(
        self,
        lookup_list: LookupKeyList,
        feature_names: List[str],
    ) -> List[Dict[str, Any]]:
        return self.lookup_engine.lookup_feature_dicts(lookup_list, feature_names)

    def cleanup(self):
        """
        Performs any cleanup associated with the online store.
        :return:
        """
        self.lookup_engine.shutdown()



import time
from sys import getsizeof
from typing import Any, Callable, Dict, List

# end-to-end latency (ms) for all lookup related logic
LOOKUP_E2E_LATENCY = "lookup_e2e_latency_ms"
# number of lookup requests made to online store
ONLINE_STORE_REQUEST_COUNT = "online_store_request_count"
# total time spent making requests to online store provider
ONLINE_STORE_REQUEST_LATENCY = "total_online_store_request_latency_ms"
# total in-memory bytes received from responses of online store provider, upper bound on wire bytes
RESPONE_IN_MEMORY_BYTES = "total_response_in_memory_bytes"
# number of features with NaN values for request
NAN_FEATURE_COUNT = "nan_feature_count"
# number of overriden feature values for request
OVERRIDEN_FEATURE_COUNT = "overriden_feature_count"


# TODO[ML-60193]: clean up this class since we start using metrics recorder defined
# in feature_store/metrics/*
class LookupClientMetrics:
    # metrics_dict - Dicitonary mapping metric name to values
    def __init__(self, metrics_dict: Dict[str, Any] = {}):
        self._metrics_dict = metrics_dict

    # increase value of metric, default 0 if no existing values for metrics
    def increase_metric(self, metric_name: str, amount: float):
        self._metrics_dict[metric_name] = (
            self._metrics_dict.get(metric_name, 0) + amount
        )

    def record_e2e_latency(self, request_duration_ms: float):
        self._metrics_dict[LOOKUP_E2E_LATENCY] = request_duration_ms

    # add request latency, bytes, and count to total combined value
    def record_request_metrics(
        self, request_duration_ms: float, in_memory_bytes: int, attempts: int = 1
    ):
        self.increase_metric(ONLINE_STORE_REQUEST_LATENCY, request_duration_ms)
        self.increase_metric(RESPONE_IN_MEMORY_BYTES, in_memory_bytes)
        self.increase_metric(ONLINE_STORE_REQUEST_COUNT, attempts)

    def get_metrics(self):
        return self._metrics_dict


# A wrapper to make patching in unittests easier.
def _current_time():
    return time.time()


def lookup_call_maybe_with_metrics(
    lookup_fn: Callable,
    metrics: LookupClientMetrics,
    measuring_e2e_latency: bool = False,
):
    if not metrics:
        return lookup_fn
    else:

        def lookup_fn_wrapper(*args, **kwargs):
            start_time = _current_time()
            online_store_output = lookup_fn(*args, **kwargs)
            duration_ms = (_current_time() - start_time) * 1000
            if measuring_e2e_latency:
                metrics.record_e2e_latency(duration_ms)
            else:
                # accumulates latency, response in-memory bytes, and count for each lookup request
                metrics.record_request_metrics(
                    duration_ms, getsizeof(online_store_output)
                )
            return online_store_output

        return lookup_fn_wrapper


def num_missing_feature_values(
    expected_feature_names: List[str], lookup_result: Dict[str, any]
):
    keys = lookup_result.keys()
    return len([f for f in expected_feature_names if f not in keys])



from typing import Any

from databricks.ml_features_common.entities._proto_enum_entity import _ProtoEnumEntity
from databricks.ml_features_common.protos.feature_store_serving_pb2 import (
    DataType as ProtoDataType,
)


class DataType(_ProtoEnumEntity):
    """Online store data types."""

    INTEGER = ProtoDataType.Value("INTEGER")
    FLOAT = ProtoDataType.Value("FLOAT")
    BOOLEAN = ProtoDataType.Value("BOOLEAN")
    STRING = ProtoDataType.Value("STRING")
    DOUBLE = ProtoDataType.Value("DOUBLE")
    LONG = ProtoDataType.Value("LONG")
    TIMESTAMP = ProtoDataType.Value("TIMESTAMP")
    DATE = ProtoDataType.Value("DATE")
    SHORT = ProtoDataType.Value("SHORT")
    ARRAY = ProtoDataType.Value("ARRAY")
    MAP = ProtoDataType.Value("MAP")
    BINARY = ProtoDataType.Value("BINARY")
    DECIMAL = ProtoDataType.Value("DECIMAL")
    STRUCT = ProtoDataType.Value("STRUCT")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoDataType



from typing import Any, Dict, List

from typing_extensions import override

# Currently the following classes are defined in the fe-client folder. It can only be used in the fe-client.
# Meaning this file cannot be imported in the LookupClient.
from databricks.ml_features.entities.data_source import DataSource
from databricks.ml_features.entities.feature import Feature
from databricks.ml_features_common.entities._feature_spec_base import _FeatureSpecBase
from databricks.ml_features_common.entities.column_info import ColumnInfo
from databricks.ml_features_common.entities.feature_spec_constants import (
    DATA_SOURCES_FIELD_NAME,
    DATA_TYPE,
    FEATURE_STORE_CLIENT_VERSION_FIELD_NAME,
    FEATURES_FIELD_NAME,
    INCLUDE,
    INPUT_COLUMNS,
    SERIALIZATION_VERSION,
    SOURCE,
    TRAINING_DATA,
    WORKSPACE_ID_FIELD_NAME,
)
from databricks.ml_features_common.entities.source_data_column_info import (
    SourceDataColumnInfo,
)


def _get_source_data_column_info_dict(
    column_infos: List[ColumnInfo],
) -> List[Dict[str, Dict[str, Any]]]:
    # TODO[ML-60159]: deprecate SOURCE and DATA_TYPE from serialization.
    return [
        {
            column_info.info.name: {
                SOURCE: TRAINING_DATA,
                INCLUDE: column_info.include,
                DATA_TYPE: column_info.data_type,
            }
        }
        for column_info in column_infos
        if column_info.info is not None
    ]


def _get_column_infos_from_dict(spec_dict: Dict[str, Any]) -> List[ColumnInfo]:
    column_infos = []
    if INPUT_COLUMNS in spec_dict:
        for column_dict in spec_dict[INPUT_COLUMNS]:
            for column_name, column_spec in column_dict.items():
                if column_spec.get(SOURCE) == TRAINING_DATA:
                    column_info = ColumnInfo(
                        info=SourceDataColumnInfo(column_name),
                        include=column_spec.get(INCLUDE, True),
                        data_type=column_spec.get(DATA_TYPE),
                    )
                    column_infos.append(column_info)
    return column_infos


class FeatureSpecDeclarative(_FeatureSpecBase):
    """
    FeatureSpecDeclarative contains a group of declarative features.
    """

    def __init__(
        self,
        *,  # Force all arguments to be keyword-only
        features: List[Feature],
        column_infos: List[ColumnInfo],
        workspace_id: int,
        feature_store_client_version: str,
        serialization_version: int,
    ):
        super().__init__(
            serialization_version, workspace_id, feature_store_client_version
        )
        self._features = features
        self._column_infos = column_infos

    @property
    def column_infos(self):
        return self._column_infos

    @property
    def features(self):
        return self._features

    def _get_data_sources(self):
        return [feature.source for feature in self._features]

    @override
    def _to_dict(self):
        """
        Deprecated[ML-59811]: FeatureSpecDeclarative should be converted to
        _SerializableFeatureSpecDeclarative before serialization.
        """
        yaml_dict = {}
        if self._column_infos:
            yaml_dict[INPUT_COLUMNS] = _get_source_data_column_info_dict(
                self._column_infos
            )
        yaml_dict[FEATURES_FIELD_NAME] = {
            feature.full_name: feature._to_yaml_dict() for feature in self._features
        }
        yaml_dict[DATA_SOURCES_FIELD_NAME] = {
            data_source.full_name(): data_source._to_yaml_dict()
            for data_source in self._get_data_sources()
        }
        # For readability, place SERIALIZATION_VERSION_NUMBER last in the dictionary.
        yaml_dict[SERIALIZATION_VERSION] = self.serialization_version
        yaml_dict[WORKSPACE_ID_FIELD_NAME] = self._workspace_id
        yaml_dict[
            FEATURE_STORE_CLIENT_VERSION_FIELD_NAME
        ] = self._feature_store_client_version
        return yaml_dict

    @classmethod
    def _from_dict(cls, spec_dict):
        """Create a FeatureSpecDeclarative from a dictionary."""
        # Parse data sources first
        data_sources = {}
        for data_source_name, data_source_dict in spec_dict[
            DATA_SOURCES_FIELD_NAME
        ].items():
            data_sources[data_source_name] = DataSource._from_yaml_dict(
                data_source_dict
            )

        # Parse features, referencing the data sources by name
        features = []
        for feature_name, feature_dict in spec_dict[FEATURES_FIELD_NAME].items():
            feature = Feature._from_yaml_dict(
                feature_name,
                feature_dict,
                data_sources[feature_dict[Feature.DATA_SOURCE_FIELD_NAME]],
            )
            features.append(feature)

        # Get serialization version
        serialization_version = spec_dict[SERIALIZATION_VERSION]

        # Parse column_infos if present
        column_infos = _get_column_infos_from_dict(spec_dict)

        # Get workspace_id and feature_store_client_version if present
        workspace_id = spec_dict.get(WORKSPACE_ID_FIELD_NAME)
        feature_store_client_version = spec_dict.get(
            FEATURE_STORE_CLIENT_VERSION_FIELD_NAME
        )

        return cls(
            column_infos=column_infos,
            features=features,
            serialization_version=serialization_version,
            workspace_id=workspace_id,
            feature_store_client_version=feature_store_client_version,
        )


class _SerializableFeatureSpecDeclarative(_FeatureSpecBase):
    """
    A representation of a FeatureSpecDeclarative that can be serialized to a YAML file.

    This format stores the minimal set of fields that allow reconstruction of full feature spec;
    this minimal set is the UC declarative feature names and input columns. It can be
    inflated to a full FeatureSpecDeclarative object by feature_spec_utils.inflate_feature_spec_declarative.

    Data source fields are not serialized since they can be resolved from the feature in UC.
    Input columns are specific to training set semantics and cannot be rederived from the features alone.

    Example of a serialized Feature Spec with declarative features:

    input_columns:
    - user_id:
        source: training_data
        include: false
        data_type: int
    - transaction_time:
        source: training_data
        include: false
        data_type: timestamp
    - amount:
        source: training_data
        include: true
        data_type: double
    features:
    - ml.dev.avg_amount_7d
    - ml.dev.avg_amount_30d_7d_slide
    - ml.dev.user_id_count_continuous_1d
    workspace_id: 6051921418418893
    feature_store_client_version: 0.13.0.dev0
    serialization_version: 11
    """

    def __init__(
        self,
        input_columns: List[ColumnInfo],
        feature_names: List[str],
        workspace_id: int,
        feature_store_client_version: str,
        serialization_version: int,
    ):
        super().__init__(
            serialization_version, workspace_id, feature_store_client_version
        )
        self._input_columns = input_columns
        self._feature_names = feature_names

    @classmethod
    def _from_feature_spec_declarative(
        cls, feature_spec_declarative: FeatureSpecDeclarative
    ):
        return cls(
            input_columns=feature_spec_declarative.column_infos,
            feature_names=[
                feature.full_name for feature in feature_spec_declarative.features
            ],
            workspace_id=feature_spec_declarative.workspace_id,
            feature_store_client_version=feature_spec_declarative.feature_store_client_version,
            serialization_version=feature_spec_declarative.serialization_version,
        )

    @override
    def _to_dict(self):
        return {
            INPUT_COLUMNS: _get_source_data_column_info_dict(self._input_columns),
            FEATURES_FIELD_NAME: self._feature_names,
            WORKSPACE_ID_FIELD_NAME: self._workspace_id,
            FEATURE_STORE_CLIENT_VERSION_FIELD_NAME: self._feature_store_client_version,
            SERIALIZATION_VERSION: self.serialization_version,
        }

    @classmethod
    def _from_dict(cls, spec_dict):
        features_from_yaml = spec_dict[FEATURES_FIELD_NAME]
        if not isinstance(features_from_yaml, list):
            # The features field is expected to be a list of feature names.
            raise ValueError(f"Invalid features format: {features_from_yaml}")
        features_names = features_from_yaml
        return cls(
            input_columns=_get_column_infos_from_dict(spec_dict),
            feature_names=features_names,
            workspace_id=spec_dict[WORKSPACE_ID_FIELD_NAME],
            feature_store_client_version=spec_dict[
                FEATURE_STORE_CLIENT_VERSION_FIELD_NAME
            ],
            serialization_version=spec_dict[SERIALIZATION_VERSION],
        )



from databricks.ml_features_common.entities._feature_store_object import (
    _FeatureStoreObject,
)
from databricks.ml_features_common.protos.feature_spec_pb2 import (
    FunctionInfo as ProtoFunctionInfo,
)


class FunctionInfo(_FeatureStoreObject):
    def __init__(self, udf_name: str):
        if not udf_name:
            raise ValueError("udf_name must be non-empty.")
        self._udf_name = udf_name

    @property
    def udf_name(self) -> str:
        return self._udf_name

    @classmethod
    def from_proto(cls, function_info_proto):
        return cls(udf_name=function_info_proto.udf_name)

    def to_proto(self):
        return ProtoFunctionInfo(udf_name=self.udf_name)



from typing import List, Tuple

import pandas as pd
import pyarrow as pa

from databricks.ml_features_common.entities.data_type import DataType

# Derived from the Spark to Arrow data type mapping in runtime.
# https://src.dev.databricks.com/databricks/runtime@78d071a6bb90371c937490f81f50f969ca020450/-/blob/sql/catalyst/src/main/scala/org/apache/spark/sql/util/ArrowUtils.scala?L39:3
BASIC_DATA_TYPE_CONVERTERS = {
    DataType.BOOLEAN: pa.bool_(),
    DataType.SHORT: pa.int16(),
    DataType.INTEGER: pa.int32(),
    DataType.LONG: pa.int64(),
    DataType.FLOAT: pa.float32(),
    DataType.DOUBLE: pa.float64(),
    DataType.STRING: pa.utf8(),  # alias for pa.string()
    DataType.BINARY: pa.binary(),
    DataType.DATE: pa.date32(),  # days since epoch
    DataType.TIMESTAMP: pa.timestamp(
        "ns"
    ),  # nanoseconds, time zone naive (implicit UTC)
}


def _spark_to_pyarrow_schema(spark_schema: List[Tuple[str, DataType]]) -> pa.Schema:
    """
    Converts a Spark schema to pyarrow schema.

    Note: Schemas with complex data types are not currently supported.
    """

    def to_pyarrow_field(col: str, spark_data_type: DataType) -> pa.Field:
        if spark_data_type in BASIC_DATA_TYPE_CONVERTERS:
            pa_data_type = BASIC_DATA_TYPE_CONVERTERS[spark_data_type]
        else:
            raise ValueError(
                f"Found unsupported data type '{DataType.to_string(spark_data_type)}' for column '{col}'."
            )
        return pa.field(col, pa_data_type)

    pa_fields = [to_pyarrow_field(col, data_type) for col, data_type in spark_schema]
    return pa.schema(pa_fields)


def cast_pandas_df(
    df: pd.DataFrame, spark_schema: List[Tuple[str, DataType]]
) -> pd.DataFrame:
    """
    Converts a pandas DataFrame to pyarrow Table for data type casting, then returns the casted DataFrame.

    Code is derived from, and should be kept in sync with Safe Spark Python UDF execution.
    https://src.dev.databricks.com/databricks/runtime@78d071a6bb90371c937490f81f50f969ca020450/-/blob/safespark/udf/py/udfserver/server.py?L204:5
    """
    result_schema = _spark_to_pyarrow_schema(spark_schema)
    # Pyarrow conversion drops the index, so we preserve it.
    index = df.index
    try:
        # Python values are converted to Arrow based on result_schema.
        # This doesn't include type coercion and any type mismatch would result in an ArrowTypeError or ArrowInvalid error.
        # Note: Newer pyarrow versions throw ArrowTypeError here, so ArrowInvalid is not required.
        # We pass `preserve_index=False` above because PyArrow otherwise stores the index as a column.
        casted_table = pa.Table.from_pandas(
            df, schema=result_schema, preserve_index=False
        )
    except (pa.ArrowTypeError, pa.ArrowInvalid):
        # In case of type mismatch, try pa.Table.cast for type coercion.
        # For example, this would allow converting string into numbers.
        casted_table = pa.Table.from_pandas(df, preserve_index=False).cast(
            target_schema=result_schema
        )

    return casted_table.to_pandas().set_index(index)



import copy
import re
from typing import Optional, Set

from databricks.ml_features_common.entities.feature_spec import FeatureSpec

SINGLE_LEVEL_NAMESPACE_REGEX = r"^[^\. \/\x00-\x1F\x7F]+$"
TWO_LEVEL_NAMESPACE_REGEX = r"^[^\. \/\x00-\x1F\x7F]+(\.[^\. \/\x00-\x1F\x7F]+)$"
THREE_LEVEL_NAMESPACE_REGEX = (
    r"^[^\. \/\x00-\x1F\x7F]+(\.[^\. \/\x00-\x1F\x7F]+)(\.[^\. \/\x00-\x1F\x7F]+)$"
)

HIVE_METASTORE_NAME = "hive_metastore"
# these two catalog names both points to the workspace local default HMS (hive metastore).
LOCAL_METASTORE_NAMES = [HIVE_METASTORE_NAME, "spark_catalog"]


# Get full table name in the form of <catalog_name>.<schema_name>.<table_name>
# given user specified table name, current catalog and schema.
def get_full_table_name(
    table_name: str,
    current_catalog: str,
    current_schema: str,
) -> str:
    _check_qualified_table_names({table_name})
    return _get_full_name_for_entity(
        name=table_name,
        current_catalog=current_catalog,
        current_schema=current_schema,
        entity_type="table",
    )


# Get full UDF name in the form of <catalog_name>.<schema_name>.<udf_name>
# given user specified UDF name, current catalog and schema.
def get_full_udf_name(
    udf_name: str,
    current_catalog: str,
    current_schema: str,
) -> str:
    _check_qualified_udf_names({udf_name})
    return _get_full_name_for_entity(
        name=udf_name,
        current_catalog=current_catalog,
        current_schema=current_schema,
        entity_type="UDF",
    )


def get_catalog_from_full_table_name(full_table_name: str) -> str:
    if not _is_three_level_name(full_table_name):
        raise _invalid_names_error({full_table_name}, "table")
    catalog, _, _ = full_table_name.split(".")
    return catalog


def _get_full_name_for_entity(
    name: str,
    current_catalog: str,
    current_schema: str,
    entity_type: str,
) -> str:
    if not _is_single_level_name(current_catalog) or not _is_single_level_name(
        current_schema
    ):
        raise ValueError(
            f"Invalid catalog '{current_catalog}' or "
            f"schema '{current_schema}' name for {entity_type} '{name}'."
        )
    if _is_single_level_name(name):
        full_name = f"{current_catalog}.{current_schema}.{name}"
    elif _is_two_level_name(name):
        full_name = f"{current_catalog}.{name}"
    elif _is_three_level_name(name):
        full_name = name
    else:
        raise _invalid_names_error({name}, entity_type)

    catalog, schema, name = full_name.split(".")
    if catalog in LOCAL_METASTORE_NAMES:
        return f"{HIVE_METASTORE_NAME}.{schema}.{name}"
    return full_name


def _replace_catalog_name(full_name: str, catalog: Optional[str]) -> str:
    if catalog is None:
        return full_name
    name_sec = full_name.split(".")
    name_sec[0] = catalog
    return ".".join(name_sec)


# Local metastore tables in feature_spec.yaml are all stored in 2L.
# Standardize table names to be all in 3L to avoid erroneously reading data from UC tables.
def get_feature_spec_with_full_table_names(
    feature_spec: FeatureSpec, catalog_name_override: Optional[str] = None
) -> FeatureSpec:
    column_info_table_names = [
        column_info.table_name for column_info in feature_spec.feature_column_infos
    ]
    table_info_table_names = [
        table_info.table_name for table_info in feature_spec.table_infos
    ]
    _check_qualified_table_names(set(column_info_table_names))
    _check_qualified_table_names(set(table_info_table_names))
    invalid_table_names = list(
        filter(_is_single_level_name, column_info_table_names)
    ) + list(filter(_is_single_level_name, table_info_table_names))
    if len(invalid_table_names) > 0:
        raise _invalid_names_error(set(invalid_table_names), "table")
    standardized_feature_spec = copy.deepcopy(feature_spec)
    for column_info in standardized_feature_spec.feature_column_infos:
        if _is_two_level_name(column_info.table_name):
            column_info._table_name = f"{HIVE_METASTORE_NAME}.{column_info.table_name}"
        column_info._table_name = _replace_catalog_name(
            column_info.table_name, catalog_name_override
        )
    for column_info in standardized_feature_spec.on_demand_column_infos:
        if _is_two_level_name(column_info.udf_name):
            column_info._udf_name = f"{HIVE_METASTORE_NAME}.{column_info.udf_name}"
        column_info._udf_name = _replace_catalog_name(
            column_info.udf_name, catalog_name_override
        )
    for table_info in standardized_feature_spec.table_infos:
        if _is_two_level_name(table_info.table_name):
            table_info._table_name = f"{HIVE_METASTORE_NAME}.{table_info.table_name}"
        table_info._table_name = _replace_catalog_name(
            table_info.table_name, catalog_name_override
        )
    for udf_info in standardized_feature_spec.function_infos:
        udf_info._udf_name = _replace_catalog_name(
            udf_info.udf_name, catalog_name_override
        )
    return standardized_feature_spec


# Reformat 3L table name for tables in local metastore to 2L. This is used when interacting with catalog client
# and serializing workspace local feature spec for scoring.
def reformat_full_table_name(full_table_name: str) -> str:
    if not _is_three_level_name(full_table_name):
        raise _invalid_names_error({full_table_name}, "table")
    catalog, schema, table = full_table_name.split(".")
    if catalog in LOCAL_METASTORE_NAMES:
        return f"{schema}.{table}"
    return full_table_name


# Reformat table names in feature_spec with reformat_full_table_name
def get_feature_spec_with_reformat_full_table_names(
    feature_spec: FeatureSpec,
) -> FeatureSpec:
    column_info_table_names = [
        column_info.table_name for column_info in feature_spec.feature_column_infos
    ]
    table_info_table_names = [
        table_info.table_name for table_info in feature_spec.table_infos
    ]
    _check_qualified_table_names(set(column_info_table_names))
    _check_qualified_table_names(set(table_info_table_names))
    invalid_table_names = list(
        filter(lambda name: not _is_three_level_name(name), column_info_table_names)
    ) + list(
        filter(lambda name: not _is_three_level_name(name), table_info_table_names)
    )
    if len(invalid_table_names) > 0:
        raise _invalid_names_error(set(invalid_table_names), "table")
    standardized_feature_spec = copy.deepcopy(feature_spec)
    for column_info in standardized_feature_spec.feature_column_infos:
        column_info._table_name = reformat_full_table_name(column_info.table_name)
    for table_info in standardized_feature_spec.table_infos:
        table_info._table_name = reformat_full_table_name(table_info.table_name)
    return standardized_feature_spec


def _invalid_names_error(invalid_names: Set[str], entity_type: str) -> ValueError:
    return ValueError(
        f"Invalid {entity_type} name{'s' if len(invalid_names) > 1 else ''} '{', '.join(invalid_names)}'."
    )


def _is_qualified_entity_name(name) -> bool:
    return isinstance(name, str) and (
        _is_single_level_name(name)
        or _is_two_level_name(name)
        or _is_three_level_name(name)
    )


def _is_single_level_name(name) -> bool:
    return (
        isinstance(name, str)
        and re.match(SINGLE_LEVEL_NAMESPACE_REGEX, name) is not None
    )


def _is_two_level_name(name) -> bool:
    return (
        isinstance(name, str) and re.match(TWO_LEVEL_NAMESPACE_REGEX, name) is not None
    )


def _is_three_level_name(name) -> bool:
    return (
        isinstance(name, str)
        and re.match(THREE_LEVEL_NAMESPACE_REGEX, name) is not None
    )


def validate_qualified_feature_name(catalog_name: str, schema_name: str, name: str):
    """
    Validates a feature name against provided catalog and schema names.

    :param catalog_name: The expected catalog name
    :param schema_name: The expected schema name
    :param name: The feature name to validate
    :raises ValueError: If validation fails with detailed error message
    """
    if not isinstance(name, str) or not name.strip():
        raise ValueError("'name' must be a non-empty string when provided")

    name = name.strip()

    # Check if name is qualified and validate catalog/schema match
    if _is_qualified_entity_name(name):
        name_parts = name.split(".")
        if len(name_parts) == 3:
            # 3-level qualified name: catalog.schema.name
            provided_catalog, provided_schema, _ = name_parts
            if provided_catalog != catalog_name:
                raise ValueError(
                    f"Catalog name mismatch: provided catalog_name='{catalog_name}' "
                    f"but name='{name}' contains catalog '{provided_catalog}'"
                )
            if provided_schema != schema_name:
                raise ValueError(
                    f"Schema name mismatch: provided schema_name='{schema_name}' "
                    f"but name='{name}' contains schema '{provided_schema}'"
                )
        elif len(name_parts) == 2:
            # 2-level qualified name not allowed with explicit catalog/schema
            raise ValueError(
                f"When catalog_name and schema_name are provided, name must be either "
                f"unqualified (single name) or fully qualified (catalog.schema.name). "
                f"Got 2-level name: '{name}'"
            )
        # For single level names, no additional validation needed


def unsupported_api_error_uc(api_name):
    return ValueError(f"{api_name} is not supported for Unity Catalog tables.")


# check if entity is in UC
def is_uc_entity(full_entity_name) -> bool:
    catalog_name, schema_name, table_name = full_entity_name.split(".")
    return not is_default_hms_table(full_entity_name)


def is_default_hms_table(full_table_name) -> bool:
    catalog_name, schema_name, table_name = full_table_name.split(".")
    return catalog_name in LOCAL_METASTORE_NAMES


# check if UDF names are in the correct format - 1L, 2L or 3L
def _check_qualified_udf_names(udf_names: Set[str]):
    unqualified_udf_names = [
        udf_name for udf_name in udf_names if not _is_qualified_entity_name(udf_name)
    ]
    if len(unqualified_udf_names) > 0:
        raise ValueError(
            f"UDF name{'s' if len(unqualified_udf_names) > 1 else ''} "
            f"'{', '.join(map(str, unqualified_udf_names))}' must have the form "
            f"<catalog_name>.<schema_name>.<udf_name>, <schema_name>.<udf_name>, "
            f"or <udf_name> and cannot include space or forward-slash."
        )


# check if table names are in the correct format - 1L, 2L or 3L
def _check_qualified_table_names(feature_table_names: Set[str]):
    unqualified_table_names = list(
        filter(
            lambda table_name: not _is_qualified_entity_name(table_name),
            feature_table_names,
        )
    )
    if len(unqualified_table_names) > 0:
        raise ValueError(
            f"Feature table name{'s' if len(unqualified_table_names) > 1 else ''} "
            f"'{', '.join(map(str, unqualified_table_names))}' must have the form "
            f"<catalog_name>.<schema_name>.<table_name>, <database_name>.<table_name>, "
            f"or <table_name> and cannot include space or forward-slash."
        )


# For APIs like create_training_set and score_batch, all tables must all be in
# UC catalog (shareable cross-workspaces) or default HMS (intended to only be used in the current workspace)
# check if all tables are either in UC or default HMS.
def _verify_all_tables_are_either_in_uc_or_in_hms(
    table_names: Set[str], current_catalog: str, current_schema: str
):
    full_table_names = [
        get_full_table_name(table_name, current_catalog, current_schema)
        for table_name in table_names
    ]
    is_valid = all(
        [is_uc_entity(full_table_name) for full_table_name in full_table_names]
    ) or all(
        [is_default_hms_table(full_table_name) for full_table_name in full_table_names]
    )
    if not is_valid:
        raise ValueError(
            f"Feature table names '{', '.join(table_names)}' "
            f"must all be in UC or the local default hive metastore. "
            f"Mixing feature tables from two different storage locations is not allowed."
        )


# For APIs like create_training_set with FeatureFunctions, only UC UDFs are supported.
def _verify_all_udfs_in_uc(
    udf_names: Set[str], current_catalog: str, current_schema: str
):
    full_udf_names = [
        get_full_udf_name(udf_name, current_catalog, current_schema)
        for udf_name in udf_names
    ]
    is_valid = all([is_uc_entity(full_udf_name) for full_udf_name in full_udf_names])
    if not is_valid:
        raise ValueError(f"UDFs must all be in Unity Catalog.")



from typing import Dict, List, Tuple

from databricks.feature_store.utils.lookup_key_utils import LookupKeyType
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.entities.feature_column_info import FeatureColumnInfo
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)


class FeatureTableMetadata:
    """
    Encapsulates metadata on a feature table, including lookup keys, feature metadata, and online
    feature table metadata.

    Feature metadata is grouped by lookup key, since different features may require different
    lookup keys (eg pickup_zip and dropoff_zip may each be used to lookup a geographic data feature
    table).
    """

    def __init__(
        self,
        feature_col_infos_by_lookup_key: Dict[
            str, Dict[LookupKeyType, List[FeatureColumnInfo]]
        ],
        online_ft: OnlineFeatureTable,
    ):
        self.feature_col_infos_by_lookup_key = feature_col_infos_by_lookup_key
        self.online_ft = online_ft

    def get_table_feature_type_mapping(self) -> Dict[str, DataType]:
        """
        Returns a mapping from feature name to data type.
        """
        feature_data_type_mapping = {}
        for feature_detail in self.online_ft.features:
            feature_data_type_mapping[feature_detail.name] = feature_detail.data_type
        return feature_data_type_mapping



import codecs
import os
from typing import Any, Dict, Optional

from yaml import dump, load

try:
    from yaml import CSafeDumper as SafeDumper
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeDumper, SafeLoader


ENCODING = "utf-8"
SAFE_DUMPER = SafeDumper


def read_yaml(path: str, file_name: Optional[str] = None) -> Dict[str, Any]:
    """
    Reads a YAML file and returns the data as a dictionary.
    :param path: Path to the YAML file.
    :param file_name: Name of the YAML file. If None, it assume the path param
    :return: Dictionary containing the data from the YAML file.
    """
    full_file_name = os.path.join(path, file_name) if file_name else path
    with open(full_file_name, "r", encoding=ENCODING) as file:
        data = load(file, Loader=SafeLoader)
    return data


def write_yaml(path: str, file_name: str, data_dict: Dict[str, Any]):
    """
    Writes a dictionary to a YAML file.
    :param path: Path to the YAML file.
    :param file_name: Name of the YAML file.
    :param data_dict: Dictionary containing the data to write.
    """
    full_file_name = os.path.join(path, file_name)
    with codecs.open(full_file_name, mode="w", encoding=ENCODING) as yaml_file:
        dump(
            data_dict,
            yaml_file,
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=False,
            Dumper=SafeDumper,
        )



"""
Utils for collection classes and functions.
"""

from typing import Any, Dict, List, Tuple


class LookupKeyList:
    """
    Lightweight collection for storing a table of lookup primary keys.
    This is faster than pandas DataFrame for small tables.
    """

    def __init__(self, columns: List[str], rows: List[List[Any]]):
        """
        Note: do not modify the columns or rows.

        :param columns: Ordered list of column names.
        :param rows: A list of rows where each row is a list of values, ordered by the columns.
        """
        self._columns = columns
        self._rows = rows
        self._items = [list(zip(columns, row)) for row in rows]

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        """
        :param d: A dictionary mapping column names to lists of values.
        :return: A LookupKeyList instance.
        """
        return cls(columns=list(d.keys()), rows=list(zip(*d.values())))

    @property
    def columns(self) -> List[str]:
        return self._columns

    @property
    def rows(self) -> List[List[Any]]:
        return self._rows

    @property
    def items(self) -> List[List[Tuple[str, Any]]]:
        """
        :return: A list of rows where each row is a list of (column, value) tuples.
        """
        return self._items

    def __eq__(self, other):
        if not isinstance(other, LookupKeyList):
            return False

        return self.columns == other.columns and self.rows == other.rows

    def __repr__(self):
        return f"LookupKeyList(columns={self.columns}, rows={self.rows})"

    def __str__(self):
        return self.__repr__()



""" Util functions for handling Delta Sharing models.
"""

from typing import Optional

from databricks.feature_store.entities.feature_functions_for_serving import (
    FeatureFunctionsForServing,
)
from databricks.ml_features_common.entities.feature_spec import FeatureSpec
from databricks.ml_features_common.entities.feature_tables_for_serving import (
    FeatureTablesForServing,
)
from databricks.ml_features_common.utils.uc_utils import HIVE_METASTORE_NAME


def get_catalog_name_override(
    feature_spec: FeatureSpec,
    feature_tables_for_serving: Optional[FeatureTablesForServing],
    feature_functions_for_serving: Optional[FeatureFunctionsForServing],
) -> Optional[str]:
    """
    Find the overriding catalog name by comparing the table and function names between
    feature-spec and the feature tables and feature functions for serving. An override is detected
    when all the catalog names in the _for_serving fields are unified and different from the
    original feature spec. This is used to handle delta-shared model and feature specs.
    :param feature_spec: original feature loaded from feature-spec.yaml
    :param feature_tables_for_serving: loaded feature tables for serving
    :param feature_functions_for_serving: loaded feature functions for serving
    :return: The overriding catalog name. None if catalog name isn't overridden.
    """
    table_names = (
        set()
        if feature_tables_for_serving is None
        else {
            table.feature_table_name
            for table in feature_tables_for_serving.online_feature_tables
        }
    )
    function_names = (
        set()
        if feature_functions_for_serving is None
        else {
            function.full_name
            for function in feature_functions_for_serving.ff_for_serving
        }
    )

    feature_spec_table_names = {table.table_name for table in feature_spec.table_infos}
    feature_spec_udf_names = {udf.udf_name for udf in feature_spec.function_infos}

    def extract_catalog_name(name: str) -> Optional[str]:
        name_sections = name.split(".")
        if len(name_sections) == 3:
            return name_sections[0]
        elif len(name_sections) == 2:
            return HIVE_METASTORE_NAME

    catalog_names = {
        extract_catalog_name(name) for name in table_names.union(function_names)
    }
    feature_spec_catalog_names = {
        extract_catalog_name(name)
        for name in feature_spec_table_names.union(feature_spec_udf_names)
    }

    if len(catalog_names) == 1 and catalog_names != feature_spec_catalog_names:
        return catalog_names.pop()
    return None



# Field names from feature_spec.proto.
SOURCE_DATA_COLUMN_INFO = "source_data_column_info"
FEATURE_COLUMN_INFO = "feature_column_info"
ON_DEMAND_COLUMN_INFO = "on_demand_column_info"
INPUT_COLUMNS = "input_columns"
NAME = "name"
OUTPUT_NAME = "output_name"
INPUT_TABLES = "input_tables"
TABLE_NAME = "table_name"
TABLE_ID = "table_id"
SERIALIZATION_VERSION = "serialization_version"
FEATURE_STORE_CLIENT_VERSION_FIELD_NAME = "feature_store_client_version"
INPUT_FUNCTIONS = "input_functions"
INCLUDE = "include"
DATA_TYPE = "data_type"
TOPOLOGICAL_ORDERING = "topological_ordering"
UDF_NAME = "udf_name"
INPUT_BINDINGS = "input_bindings"
PARAMETER = "parameter"
BOUND_TO = "bound_to"
WORKSPACE_ID_FIELD_NAME = "workspace_id"
FEATURES_FIELD_NAME = "features"

# FeatureSpec YAML fields not defined in the proto file:
# These fields do not exist in the proto file because they only exist in
# the python logic and are not serialized.
SOURCE = "source"
TRAINING_DATA = "training_data"
FEATURE_STORE = "feature_store"
ON_DEMAND_FEATURE = "on_demand_feature"

# FeatureSpec YAML field for data sources
DATA_SOURCES_FIELD_NAME = "data_sources"



"""
Metrics recording system for Feature Store lookup client.

This module provides a pluggable metrics interface that allows the Feature Store
lookup client to emit operational metrics (latency, count, errors) to model server.

The metrics system uses a global registry pattern:
- Runtime environments (model server, serving scheduler) register their metrics
  implementation via `register_metrics_recorder()`
- The lookup client retrieves the recorder via `get_metrics_recorder()` to emit metrics
- If no recorder is registered, a no-op implementation is used (zero overhead)

Typical usage in model server:
    from databricks.feature_store.metrics import register_metrics_recorder
    register_metrics_recorder(PrometheusMetricsRecorder())

Typical usage in lookup client:
    from databricks.feature_store.metrics import get_metrics_recorder
    recorder = get_metrics_recorder()
    recorder.record_histogram("lookup_latency", latency_ms, {"is_postgres": "true"})
"""

from databricks.feature_store.metrics.feature_store_metrics_recorder import (
    get_metrics_recorder,
    register_metrics_recorder,
)

__all__ = ["register_metrics_recorder", "get_metrics_recorder"]



import abc
import pprint
from itertools import chain

from databricks.ml_features_common.utils.utils_common import is_marked_deprecated


class _FeatureStoreObject(abc.ABC):
    def __iter__(self):
        # Iterate through list of public attributes and properties and yield as key -> value
        for prop in chain(self._public_attributes(), self._properties()):
            yield prop, self.__getattribute__(prop)

    def _public_attributes(self):
        """
        :return a list of public attributes. Public attribute's name does not start with "_"
        """
        return [a for a in self.__dict__.keys() if not a.startswith("_")]

    @classmethod
    def _properties(cls):
        """
        :return a list of properties that is not marked as @deprecated
        """
        return sorted(
            [
                p
                for p in cls.__dict__
                if isinstance(getattr(cls, p), property)
                and not is_marked_deprecated(getattr(getattr(cls, p), "fget"))
            ]
        )

    @classmethod
    def from_dictionary(cls, the_dict):
        filtered_dict = {
            key: value for key, value in the_dict.items() if key in cls._properties()
        }
        return cls(**filtered_dict)

    def __repr__(self):
        return _EntityPrinter().to_string(self)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__


class _EntityPrinter(object):
    def __init__(self):
        super().__init__()
        self.printer = pprint.PrettyPrinter(indent=1)

    def to_string(self, obj):
        if isinstance(obj, _FeatureStoreObject):
            return f"<{type(obj).__name__}: {self._entity_to_string(obj)}>"
        return self.printer.pformat(obj)

    def _entity_to_string(self, entity):
        return ", ".join([f"{key}={self.to_string(value)}" for key, value in entity])



from typing import List

from databricks.ml_features_common.entities.column_info import ColumnInfo
from databricks.ml_features_common.entities.feature_column_info import FeatureColumnInfo
from databricks.ml_features_common.entities.feature_spec import FeatureSpec
from databricks.ml_features_common.entities.feature_table_info import FeatureTableInfo
from databricks.ml_features_common.entities.function_info import FunctionInfo
from databricks.ml_features_common.entities.on_demand_column_info import (
    OnDemandColumnInfo,
)

# TODO (ML-30526): Move this file to the tests module.

TEST_WORKSPACE_ID = 123


def get_test_table_infos_from_column_infos(column_infos: List[ColumnInfo]):
    table_infos = []
    unique_table_names = set(
        [
            column_info.info.table_name
            for column_info in column_infos
            if isinstance(column_info, ColumnInfo)
            and isinstance(column_info.info, FeatureColumnInfo)
        ]
    )
    for table_name in unique_table_names:
        table_id = table_name + "123456"
        table_infos.append(FeatureTableInfo(table_name=table_name, table_id=table_id))
    return table_infos


def get_test_function_infos_from_column_infos(column_infos: List[ColumnInfo]):
    function_infos = []
    unique_udf_names = set(
        [
            column_info.info.udf_name
            for column_info in column_infos
            if isinstance(column_info, ColumnInfo)
            and isinstance(column_info.info, OnDemandColumnInfo)
        ]
    )
    for udf_name in unique_udf_names:
        function_infos.append(FunctionInfo(udf_name=udf_name))
    return function_infos


def create_test_feature_spec(
    column_infos: List[ColumnInfo],
    table_infos: List[FeatureTableInfo] = None,
    function_infos: List[FunctionInfo] = None,
    workspace_id: int = TEST_WORKSPACE_ID,
):
    if table_infos is None:
        table_infos = get_test_table_infos_from_column_infos(column_infos)
    if function_infos is None:
        function_infos = get_test_function_infos_from_column_infos(column_infos)
    return FeatureSpec(
        column_infos=column_infos,
        table_infos=table_infos,
        function_infos=function_infos,
        workspace_id=workspace_id,
        feature_store_client_version="test0",
        serialization_version=FeatureSpec.SERIALIZATION_VERSION_NUMBER,
    )



import typing

from databricks.feature_store.entities.feature_functions_for_serving import (
    FeatureFunctionForServing,
    FeatureFunctionParameterInfo,
    FeatureFunctionsForServing,
)
from databricks.feature_store.feature_functions.utils.codegen_utils import CodegenUtils

# USER_FUNC_NAME is used as the function name of the Feature Function UDF in generated code. It
# is the same value used by Safe Spark when executing Python UDFs:
# https://github.com/databricks/runtime/blob/f976bbb0f71077c5f79015ffa87d930b249061ab/safespark/udf/py/udfserver/server.py#L138
# We assume that the user does not have a global value with conflicting name in their UDF
# routine_definition.
USER_FUNC_NAME = "_udf_handler_func_entry"


class FeatureFunctionLoader:
    """
    This class transforms Feature Functions defined by FeatureFunctionsForServing into Python
    functions.

    Logic to transform FeatureFunctionForServing function definition strings into Python callable
    objects is run within a separate scope from the calling environment. As such, closures are not
    supported.

    The function is executed within the same Python environment as the caller, so all packages
    installed in the Python environment are available.
    """

    @staticmethod
    def load_feature_function(
        ff_for_serving: FeatureFunctionForServing,
    ) -> typing.Callable:
        """
        Loads a single Feature Function
        """
        globals = {}
        generated_code = CodegenUtils.generate_function(USER_FUNC_NAME, ff_for_serving)
        # `exec` will execute code outside of the local + global scope, and dumps global variables
        # into `globals`. The generated code declares a global USER_FUNC_NAME equal to the Python
        # function. We can then extract this function from the `globals`
        exec(generated_code, globals)
        return globals[USER_FUNC_NAME]

    @staticmethod
    def load_feature_functions_map(
        ffs_for_serving: FeatureFunctionsForServing,
    ) -> typing.Dict[str, typing.Callable]:
        """
        Loads all FeatureFunctions included in ffs_for_serving.

        Returns a dictionary mapping function name to function.

        load_feature_functions_map(
          FeatureFunctionsForServing(
            ff_for_servign = [
              FeatureFunctionForServing("cat.sch.fn1", ...),
              FeatureFunctionForServing("cat.sch.fn2", ...),
            ]
          )
        )

        =>

        {
          "cat.sch.fn1": fn1,
          "cat.sch.fn2": fn2
        }

        Where fn1 and fn2 are functions.
        """
        return {
            ff_for_serving.full_name: FeatureFunctionLoader.load_feature_function(
                ff_for_serving
            )
            for ff_for_serving in ffs_for_serving.ff_for_serving
        }



from typing import Optional

from databricks.ml_features_common.entities._feature_store_object import (
    _FeatureStoreObject,
)
from databricks.ml_features_common.protos.feature_spec_pb2 import (
    FeatureTableInfo as ProtoFeatureTableInfo,
)


class FeatureTableInfo(_FeatureStoreObject):
    def __init__(
        self,
        table_name: str,
        table_id: Optional[str],
        lookback_window: Optional[float] = None,
    ):
        if not table_name:
            raise ValueError("table_name must be non-empty.")
        self._table_name = table_name
        self._table_id = table_id
        self._lookback_window = lookback_window

    @property
    def table_name(self):
        return self._table_name

    @property
    def table_id(self):
        return self._table_id

    @property
    def lookback_window(self):
        return self._lookback_window

    @classmethod
    def from_proto(cls, feature_table_info_proto):
        lookback_window = (
            feature_table_info_proto.lookback_window
            if feature_table_info_proto.HasField("lookback_window")
            else None
        )
        return cls(
            table_name=feature_table_info_proto.table_name,
            table_id=feature_table_info_proto.table_id,
            lookback_window=lookback_window,
        )

    def to_proto(self):
        return ProtoFeatureTableInfo(
            table_name=self.table_name,
            table_id=self.table_id,
            lookback_window=self.lookback_window,
        )



import inspect
import itertools
import logging
import os
import time
from collections import Counter, defaultdict
from typing import Any, Callable, Dict, List, Optional, Union

import mlflow
import pandas as pd
from mlflow.entities import SpanType
from mlflow.utils import databricks_utils

from databricks.feature_store.entities.feature_functions_for_serving import (
    FeatureFunctionsForServing,
)
from databricks.feature_store.entities.feature_table_metadata import (
    FeatureTableMetadata,
)
from databricks.feature_store.feature_functions.feature_function_executor import (
    FeatureFunctionExecutor,
)
from databricks.feature_store.feature_lookup_version import VERSION
from databricks.feature_store.metrics import get_metrics_recorder
from databricks.feature_store.metrics.feature_store_metrics_recorder import (
    METRIC_FEATURE_STORE_ERROR_COUNT,
    METRIC_FEATURE_STORE_LATENCY,
    METRIC_LABEL_ERROR_SOURCE,
    METRIC_LABEL_ONLINE_STORE_TYPE,
    METRIC_RAW_MODEL_PREDICT_LATENCY,
)
from databricks.feature_store.online_lookup_client import (
    OnlineLookupClient,
    can_use_brickstore_http_gateway,
    is_primary_key_lookup,
    tables_share_dynamodb_access_keys,
)
from databricks.feature_store.postgres_fetcher import PostgresFetcher
from databricks.feature_store.utils.collection_utils import LookupKeyList
from databricks.feature_store.utils.delta_sharing_utils import get_catalog_name_override
from databricks.feature_store.utils.latency_debugger import LatencyDebugger
from databricks.feature_store.utils.logging_utils import get_logger
from databricks.feature_store.utils.lookup_client_envvars import (
    ENV_SERVABLE_TYPE,
    FEATURE_SERVING_HP,
    FEATURE_TABLES_FOR_SERVING_FILEPATH_ENV,
    LOOKUP_CLIENT_FEATURE_FUNCTION_EVALUATION_ENABLED_ENV,
    LOOKUP_PERFORMANCE_DEBUG_ENABLED,
)
from databricks.feature_store.utils.lookup_key_utils import (
    LookupKeyType,
    get_primary_key_list,
)
from databricks.feature_store.utils.metrics_utils import (
    OVERRIDEN_FEATURE_COUNT,
    LookupClientMetrics,
    lookup_call_maybe_with_metrics,
)
from databricks.feature_store.utils.timeout_utils import (
    TOTAL_REQUEST_TIMEOUT_SECONDS,
    timeout,
)
from databricks.feature_store.utils.trace_utils import (
    is_feature_tracing_enabled,
    trace_latency,
)
from databricks.ml_features_common import mlflow_model_constants
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.entities.feature_column_info import FeatureColumnInfo
from databricks.ml_features_common.entities.feature_definition_type import (
    FeatureDefinitionType,
)
from databricks.ml_features_common.entities.feature_spec import FeatureSpec
from databricks.ml_features_common.entities.feature_tables_for_serving import (
    FeatureTablesForServing,
)
from databricks.ml_features_common.entities.on_demand_column_info import (
    OnDemandColumnInfo,
)
from databricks.ml_features_common.entities.online_feature_table import (
    OnlineFeatureTable,
)
from databricks.ml_features_common.entities.store_type import StoreType
from databricks.ml_features_common.mlflow_model_constants import _NO_RESULT_TYPE_PASSED
from databricks.ml_features_common.utils.data_type_utils import (
    deserialize_default_value_to_data_type,
)
from databricks.ml_features_common.utils.feature_spec_utils import (
    COLUMN_INFO_TYPE_FEATURE,
    COLUMN_INFO_TYPE_ON_DEMAND,
    COLUMN_INFO_TYPE_SOURCE,
    get_feature_execution_groups,
)
from databricks.ml_features_common.utils.uc_utils import (
    get_feature_spec_with_full_table_names,
    get_feature_spec_with_reformat_full_table_names,
)

_logger = get_logger(__name__, log_level=logging.INFO)


def _load_raw_model(path):
    raw_model_path = os.path.join(path, mlflow_model_constants.RAW_MODEL_FOLDER)
    return mlflow.pyfunc.load_model(raw_model_path)


class RawModelException(Exception):
    """Wrapper exception raised when the raw model predict fails."""

    pass


class _FeatureStoreModelWrapper:
    def __init__(
        self, path: str, current_time_seconds_fn: Callable[[], float] = time.time
    ):
        _logger.info(f"Initializing feature store lookup client: {VERSION}")
        self._check_support()
        feature_tables_for_serving = self._load_feature_tables_for_serving()
        feature_functions_for_serving = None
        self.has_feature_functions = False
        # TODO[ML-60195]: deprecate latency debugger since we have metrics recorder.
        self._latency_debugger = LatencyDebugger(current_time_seconds_fn)

        if self._is_lookup_client_feature_function_evaluation_enabled():
            self.has_feature_functions = self._has_feature_functions(path)
            if self.has_feature_functions:
                feature_function_dir = os.path.join(
                    path, mlflow_model_constants.FEATURE_STORE_INTERNAL_DATA_DIR
                )
                feature_functions_for_serving = FeatureFunctionsForServing.load(
                    feature_function_dir
                )

        self.raw_model = _load_raw_model(path)

        # Reformat local metastore 3L tables to 2L. Non-local metastore 3L tables and 2L tables are unchanged.
        # This guarantees table name consistency between feature_spec and feature_tables_for_serving.
        # https://docs.google.com/document/d/1x_V9GshlnoAAFFCuDsXWdJVtop9MG2HWTUo5IK_1mEw
        original_feature_spec = self._load_or_construct_feature_spec(
            feature_tables_for_serving, path
        )

        catalog_name_override = get_catalog_name_override(
            original_feature_spec,
            feature_tables_for_serving,
            feature_functions_for_serving,
        )

        # We call get_feature_spec_with_full_table_names to append the default metastore to 2L names,
        # as get_feature_spec_with_reformat_full_table_names expects full 3L table names and throws otherwise.
        # TODO (ML-26593): Consolidate this into a single function that allows either 2L/3L names.
        feature_spec_with_full_table_names = get_feature_spec_with_full_table_names(
            original_feature_spec, catalog_name_override
        )
        self.feature_spec = get_feature_spec_with_reformat_full_table_names(
            feature_spec_with_full_table_names
        )

        self.ft_metadata = self._get_ft_metadata(
            self.feature_spec, feature_tables_for_serving
        )
        self._validate_ft_metadata(self.ft_metadata)
        # Print the table names for users to validate.
        lakebase_tables = []
        for _, meta in self.ft_metadata.items():
            if (
                meta.online_ft.online_store.store_type
                == StoreType.DATABRICKS_ONLINE_STORE
            ):
                lakebase_tables.append(meta.online_ft.online_feature_table_name)
        if len(lakebase_tables) > 0:
            _logger.info(
                f"Tables in Databricks Online Feature Store: {lakebase_tables}"
            )
        self.eligible_for_pgsql_connection = self._eligible_for_pgsql_connection()
        self.is_model_eligible_for_batch_lookup = (
            self._is_model_eligible_for_batch_lookup(self.ft_metadata)
        )
        if (
            self.eligible_for_pgsql_connection
            or not self.is_model_eligible_for_batch_lookup
        ):
            self.ft_to_lookup_client = self._create_lookup_clients(
                self.ft_metadata, self.eligible_for_pgsql_connection
            )
        else:
            self.batch_lookup_client = self._create_batch_lookup_client(
                self.ft_metadata
            )
        if self.has_feature_functions:
            self.feature_function_executor = FeatureFunctionExecutor(
                feature_functions_for_serving,
                self.feature_spec.on_demand_column_infos,
            )
        else:
            self.feature_function_executor = None
        if self.eligible_for_pgsql_connection:
            self.pg_sql_fetcher = PostgresFetcher(
                self.feature_spec,
                self.ft_to_lookup_client,
                self.ft_metadata,
                self.feature_function_executor,
                self._latency_debugger,
            )
        self.table_feature_data_type_mapping = self._map_feature_to_data_type(
            self.ft_metadata
        )
        # The string representation of the store type of the first online table.
        # We do not support multiple store types in the same lookup client so
        # this value is representative.
        self.store_type_str = (
            StoreType.to_string(
                list(self.ft_metadata.values())[0].online_ft.online_store.store_type
            )
            if self.ft_metadata
            else "NONE"
        )

    # TODO (ML-59471): Decouple FeatureSpec from lookup client for declarative features.
    def _load_or_construct_feature_spec(
        self, feature_tables_for_serving: FeatureTablesForServing, path: str
    ) -> FeatureSpec:
        """
        Loads a FeatureSpec from disk or constructs it from declarative features.

        If the feature definition type is DECLARATIVE_FEATURES, constructs the FeatureSpec object
        from the feature_tables_for_serving data. Otherwise, loads the
        FeatureSpec from the specified path.
        """
        if (
            feature_tables_for_serving.feature_definition_type
            == FeatureDefinitionType.DECLARATIVE_FEATURES
        ):
            return FeatureSpec.from_feature_tables_for_serving(
                feature_tables_for_serving,
            )
        else:
            return FeatureSpec.load(path)

    def _map_feature_to_data_type(
        self, ft_metadatas: Dict[str, FeatureTableMetadata]
    ) -> Dict[str, Dict[str, DataType]]:
        """
        Maps feature names to their data types. This is used to deserialize the default values
        of features in the FeatureSpec.
        """
        table_feature_data_type_mapping = defaultdict(
            lambda: dict()
        )  # oft_name -> feature_name -> data_type
        for ft_name, ft_meta in ft_metadatas.items():
            oft = ft_meta.online_ft
            for feature_detail in oft.features:
                table_feature_data_type_mapping[
                    oft.online_feature_table_name
                ] = ft_meta.get_table_feature_type_mapping()
        return table_feature_data_type_mapping

    def _has_feature_functions(self, path):
        feature_function_dir = os.path.join(
            path, mlflow_model_constants.FEATURE_STORE_INTERNAL_DATA_DIR
        )
        return os.path.isfile(
            os.path.join(feature_function_dir, FeatureFunctionsForServing.DATA_FILE)
        )

    # true if all feature tables using DynamoDB under same authorization (ie same region and keys)
    def _is_model_eligible_for_batch_lookup(
        self, ft_metadata: Dict[str, FeatureTableMetadata]
    ):
        online_feature_tables = [meta.online_ft for _, meta in ft_metadata.items()]
        return (
            tables_share_dynamodb_access_keys(online_feature_tables)
            and is_primary_key_lookup(online_feature_tables)
        ) or can_use_brickstore_http_gateway(online_feature_tables)

    @staticmethod
    def _load_feature_tables_for_serving() -> FeatureTablesForServing:
        return FeatureTablesForServing.load(
            path=os.getenv(FEATURE_TABLES_FOR_SERVING_FILEPATH_ENV)
        )

    @staticmethod
    def _is_lookup_performance_debug_enabled() -> bool:
        # environment variables stored as string, rather than boolean
        return os.getenv(LOOKUP_PERFORMANCE_DEBUG_ENABLED) == "true"

    @staticmethod
    def _is_lookup_client_feature_function_evaluation_enabled() -> bool:
        return (
            os.getenv(LOOKUP_CLIENT_FEATURE_FUNCTION_EVALUATION_ENABLED_ENV) == "true"
        )

    @staticmethod
    def _check_support():
        if (
            databricks_utils.is_in_databricks_notebook()
            or databricks_utils.is_in_databricks_job()
        ):
            raise NotImplementedError(
                "Feature Store packaged models cannot be loaded with MLflow APIs. For batch "
                "inference, use FeatureStoreClient.score_batch."
            )

    @staticmethod
    def _validate_ft_metadata(ft_metadata: Dict[str, FeatureTableMetadata]):
        for ft, meta in ft_metadata.items():
            for lookup_key in meta.feature_col_infos_by_lookup_key.keys():
                if len(lookup_key) != len(meta.online_ft.primary_keys):
                    raise Exception(
                        f"Internal error: Online feature table has primary keys "
                        f"{meta.online_ft.primary_keys}, however FeatureSpec specifies "
                        f"{len(lookup_key)} lookup_keys: {lookup_key}."
                    )

    @staticmethod
    def _current_time_ms():
        # consider switching to time.perf_counter if more precise result is needed.
        return int(time.time() * 1000)

    def _create_lookup_clients(
        self, ft_metadata: Dict[str, FeatureTableMetadata], eligible_for_pgsql: bool
    ) -> Dict[str, OnlineLookupClient]:
        ft_to_lookup_client = {}
        for ft, meta in ft_metadata.items():
            ft_to_lookup_client[ft] = OnlineLookupClient(
                meta.online_ft,
                len(ft_metadata),
                eligible_for_pgsql,
            )
        return ft_to_lookup_client

    def _create_batch_lookup_client(
        self, ft_metadata: Dict[str, FeatureTableMetadata]
    ) -> Optional[OnlineLookupClient]:
        """
        Creates and returns an OnlineLookupClient for batch lookup, or None if `ft_metadata` is
        empty.
        """
        if not ft_metadata:
            return None
        online_fts = []
        for ft, meta in ft_metadata.items():
            online_fts.append(meta.online_ft)
        # Batch lookup is not supported for PGSQL connection
        return OnlineLookupClient(online_fts, None, eligible_for_pgsql=False)

    @trace_latency("feature_lookup")
    def _monitored_augment_with_materialized_features(
        self,
        df: pd.DataFrame,
        features_to_lookup: List[FeatureColumnInfo],
        partially_overridden_feature_output_names: List[str],
        *,
        output_metrics: LookupClientMetrics = None,
    ):
        augment_with_materialized_features = lookup_call_maybe_with_metrics(
            self._augment_with_materialized_features,
            output_metrics,
            measuring_e2e_latency=True,
        )

        # Add materialized features to `df`. This call does not do input column filtering, which is
        # important as Feature Functions may require inputs with include=False, so filtering must
        # be done after Feature Function execution. See _augment_with_materialized_features
        # docstring for details.
        return augment_with_materialized_features(
            df,
            features_to_lookup,
            partially_overridden_feature_output_names,
            metrics=output_metrics,
        )

    @trace_latency("feature_function")
    def _augment_with_on_demand_features(self, df, functions_to_apply):
        if (
            self._is_lookup_client_feature_function_evaluation_enabled()
            and self.has_feature_functions
        ):
            # Includes fully and partially overridden columns in this iteration.
            df_columns_set = set(df.columns)
            override_columns = [
                f.output_name
                for f in functions_to_apply
                if f.output_name in df_columns_set
            ]
            if override_columns:
                overrides = df[override_columns]
                # May contain partially overridden columns that will be used in later iterations.
                df_without_override_columns = df.drop(columns=override_columns)

            else:
                overrides = None
                df_without_override_columns = df

            feature_functions_df = self._compute_feature_functions(
                df_without_override_columns, functions_to_apply, overrides
            )
            return pd.concat(
                [df_without_override_columns, feature_functions_df], axis=1
            )
        return df

    # basic predict just returning predict
    def predict(self, df: pd.DataFrame, params: Optional[Dict[str, Any]] = None):
        try:
            return self._predict(df, params)
        except Exception as e:
            if isinstance(e, RawModelException):
                metrics_label = {METRIC_LABEL_ERROR_SOURCE: "raw_model"}
            else:
                # TODO[ML-60339]: Make error source more granular.
                metrics_label = {METRIC_LABEL_ERROR_SOURCE: "feature_server"}
            # Record the error
            get_metrics_recorder().record_counter(
                METRIC_FEATURE_STORE_ERROR_COUNT, 1, metrics_label
            )
            raise

    # new method to return tuple of prediction and metrics if available
    # TODO[ML-60193]: deprecate this once it's removed from model serving.
    def predict_with_metrics(
        self, df: pd.DataFrame, params: Optional[Dict[str, Any]] = None
    ):
        try:
            prediction = self._predict(df, params, None)
            return (prediction, None)
        except Exception as e:
            if isinstance(e, RawModelException):
                metrics_label = {METRIC_LABEL_ERROR_SOURCE: "raw_model"}
            else:
                # TODO[ML-60339]: Make error source more granular.
                metrics_label = {METRIC_LABEL_ERROR_SOURCE: "feature_lookup"}
            # Record the error
            get_metrics_recorder().record_counter(
                METRIC_FEATURE_STORE_ERROR_COUNT, 1, metrics_label
            )
            raise

    # Dispatch to the correct predict method based on conditions.
    # TODO[ML-60193]: deprecate param output_metrics.
    @trace_latency("databricks_feature_store")
    def _predict(
        self,
        df: pd.DataFrame,
        params: Optional[Dict[str, Any]] = None,
        output_metrics=None,
    ) -> Union[Dict[str, Any], pd.DataFrame]:
        if self._is_lookup_performance_debug_enabled():
            self._latency_debugger.init()

        result = {}
        if self.eligible_for_pgsql_connection:
            # Start time of the predict method. Used for latency measurement.
            predict_start_time = self._current_time_ms()
            metrics_recorder = get_metrics_recorder()
            if os.getenv(ENV_SERVABLE_TYPE) == "FEATURES":
                # Feature Serving: return the feature values directly as a Dict.
                result = self.pg_sql_fetcher.lookup_features(df)
                feature_lookup_latency = self._current_time_ms() - predict_start_time
                metrics_recorder.record_histogram(
                    METRIC_FEATURE_STORE_LATENCY,
                    feature_lookup_latency,
                    {METRIC_LABEL_ONLINE_STORE_TYPE: self.store_type_str},
                )
                # TODO[ML-60195]: deprecate latency debugger.
                self._latency_debugger.add_latency("pg_lookup_features")
            else:
                # Model Serving: fetch the feature values in DataFrame and run the model prediction.
                model_input_df = self.pg_sql_fetcher.get_model_input_df(df)
                feature_lookup_time = self._current_time_ms()
                metrics_recorder.record_histogram(
                    METRIC_FEATURE_STORE_LATENCY,
                    feature_lookup_time - predict_start_time,
                    {METRIC_LABEL_ONLINE_STORE_TYPE: self.store_type_str},
                )
                # TODO[ML-60195]: deprecate latency debugger.
                self._latency_debugger.add_latency("pg_get_model_input_df")
                result = self._raw_model_predict(model_input_df, params)
                raw_model_predict_time = self._current_time_ms()
                metrics_recorder.record_histogram(
                    METRIC_RAW_MODEL_PREDICT_LATENCY,
                    raw_model_predict_time - feature_lookup_time,
                )
                self._latency_debugger.add_latency("model_predict")
        else:
            result = self._legacy_predict(df, params, output_metrics)
        latency_debug_string = self._latency_debugger.debug_string()
        if latency_debug_string:
            _logger.info(latency_debug_string)
        return result

    @trace_latency("databricks_feature_store")
    def _legacy_predict(
        self,
        df: pd.DataFrame,
        params: Optional[Dict[str, Any]] = None,
        output_metrics=None,
    ):
        """
        This is the legacy predict method that supports
        * 3rd party online stores
        * Brickstore and Lakebase through HTTP gateway
        * DAG execution
        """
        predict_start_time = self._current_time_ms()
        metrics_recorder = get_metrics_recorder()
        (
            fully_overridden_feature_output_names,
            partially_overridden_feature_output_names,
        ) = self._get_overridden_feature_output_names(df)
        fully_overridden_feature_output_names_set = set(
            fully_overridden_feature_output_names
        )

        self._validate_input(
            df,
            fully_overridden_feature_output_names,
            partially_overridden_feature_output_names,
        )
        self._latency_debugger.add_latency("validation")

        if output_metrics and len(fully_overridden_feature_output_names) > 0:
            output_metrics.increase_metric(
                OVERRIDEN_FEATURE_COUNT,
                df[fully_overridden_feature_output_names].count().sum(),
            )
        self._latency_debugger.add_latency("m_override_ft_ctn")

        execution_groups = get_feature_execution_groups(self.feature_spec, df.columns)
        result_df = df
        self._latency_debugger.add_latency("load_groups")
        for execution_group in execution_groups:
            if execution_group.type == COLUMN_INFO_TYPE_SOURCE:
                self._latency_debugger.add_latency("eg_source")
                continue
            elif execution_group.type == COLUMN_INFO_TYPE_FEATURE:
                # Filter out fully overridden feature names, so they are not looked up unnecessarily
                result_df = self._monitored_augment_with_materialized_features(
                    result_df,
                    [
                        feature
                        for feature in execution_group.features
                        if feature.output_name
                        not in fully_overridden_feature_output_names_set
                    ],
                    partially_overridden_feature_output_names,
                    output_metrics=output_metrics,
                )
                self._latency_debugger.add_latency("eg_lookup")
            elif execution_group.type == COLUMN_INFO_TYPE_ON_DEMAND:
                result_df = self._augment_with_on_demand_features(
                    result_df, execution_group.features
                )
                self._latency_debugger.add_latency("eg_func")
            else:
                # This should never be visible to user.
                raise Exception(
                    "Unknown feature execution group type:", execution_group.type
                )
        self._latency_debugger.add_latency("execution_group_done")

        output_cols = [
            ci.output_name for ci in self.feature_spec.column_infos if ci.include
        ]
        model_input_df = result_df[output_cols]
        self._latency_debugger.add_latency("reorder_results")
        feature_lookup_time = self._current_time_ms()
        metrics_recorder.record_histogram(
            METRIC_FEATURE_STORE_LATENCY,
            feature_lookup_time - predict_start_time,
            {METRIC_LABEL_ONLINE_STORE_TYPE: self.store_type_str},
        )

        # Feature Trace logging
        # TODO: ML-49161 implement size limit for feature trace logging if needed
        if is_feature_tracing_enabled():
            with mlflow.start_span("custom_model", SpanType.RETRIEVER) as span:
                span.set_inputs(
                    {"model_input": model_input_df.to_dict(orient="records")}
                )
                prediction = self._raw_model_predict(model_input_df, params)
        else:
            prediction = self._raw_model_predict(model_input_df, params)
        raw_model_predict_time = self._current_time_ms()
        metrics_recorder.record_histogram(
            METRIC_RAW_MODEL_PREDICT_LATENCY,
            raw_model_predict_time - feature_lookup_time,
        )
        self._latency_debugger.add_latency("predict")
        return prediction

    def _raw_model_predict(
        self,
        model_input_df: pd.DataFrame,
        params: Optional[Dict[str, Any]] = None,
    ):
        try:
            if inspect.signature(self.raw_model.predict).parameters.get("params"):
                # Avoid passing result_type as params if it has the default value.
                # We are almost certain this isn't what the raw model expects.
                if (
                    params
                    and params.get("result_type", _NO_RESULT_TYPE_PASSED)
                    == _NO_RESULT_TYPE_PASSED
                ):
                    params.pop("result_type", None)
                prediction = self.raw_model.predict(model_input_df, params=params)
            else:
                prediction = self.raw_model.predict(model_input_df)
                if params:
                    _logger.warning(
                        "The model does not support passing additional "
                        "parameters to the predict function. Supplied parameters "
                        "will be ignored."
                    )
            return prediction
        except Exception as e:
            # Raise a custom exception so we can record the error type and message.
            raise RawModelException from e

    def _get_ft_metadata(
        self,
        feature_spec: FeatureSpec,
        fts_for_serving: FeatureTablesForServing,
    ) -> Dict[str, FeatureTableMetadata]:
        ft_to_lookup_key_to_feature_col_infos = (
            self._group_fcis_by_feature_table_lookup_key(feature_spec)
        )
        ft_names = ft_to_lookup_key_to_feature_col_infos.keys()
        ft_to_online_ft = self._resolve_online_stores(
            feature_tables=ft_names, feature_tables_for_serving=fts_for_serving
        )

        return {
            ft: FeatureTableMetadata(
                feature_col_infos_by_lookup_key=ft_to_lookup_key_to_feature_col_infos[
                    ft
                ],
                online_ft=ft_to_online_ft[ft],
            )
            for ft in ft_names
        }

    @staticmethod
    def _group_fcis_by_feature_table_lookup_key(
        feature_spec: FeatureSpec,
    ) -> Dict[str, Dict[LookupKeyType, List[FeatureColumnInfo]]]:
        """
        Re-organizes the provided FeatureSpec into a convenient structure for creating
        FeatureTableMetadata objects.

        :return: Nested dictionary:
            {feature_table_name -> {lookup_key -> feature_column_infos}}
        """
        feature_table_to_lookup_key_to_fcis = defaultdict(lambda: defaultdict(list))
        for fci in feature_spec.feature_column_infos:
            feature_table_name = fci.table_name
            lookup_key = tuple(fci.lookup_key)
            feature_table_to_lookup_key_to_fcis[feature_table_name][lookup_key].append(
                fci
            )
        return feature_table_to_lookup_key_to_fcis

    def _get_overridden_feature_output_names(self, df: pd.DataFrame):
        """A feature value can be overridden in the provided DataFrame. There are two cases:
          1. The feature value is overridden for all rows in the df.
          2. The feature value is overridden for some but not all rows in the df.

        :return: Tuple<List[str], List[str]>
          (List of feature names with values that were fully overridden,
          list of feature names with values that were partially overridden)
        """
        df_columns_set = set(df.columns)
        overridden_feature_names = [
            fci.output_name
            for fci in self.feature_spec.feature_column_infos
            if fci.output_name in df_columns_set
        ]
        if len(overridden_feature_names) == 0:
            return [], []

        all_rows_overridden_idxs = df[overridden_feature_names].notna().all().values
        fully_overridden = [
            overridden_feature_names[i]
            for i in range(len(overridden_feature_names))
            if all_rows_overridden_idxs[i]
        ]
        fully_overridden_set = set(fully_overridden)
        partially_overridden = [
            name
            for name in overridden_feature_names
            if name not in fully_overridden_set
        ]
        return (fully_overridden, partially_overridden)

    def _eligible_for_pgsql_connection(self):
        if os.getenv(FEATURE_SERVING_HP) == "false":
            return False
        online_stores = [
            ft_metadata.online_ft.online_store
            for ft_metadata in self.ft_metadata.values()
        ]
        # note: it's redundant to check the length of online_stores. But adding this for clarity.
        if len(online_stores) > 0 and not all(
            online_store.store_type is StoreType.DATABRICKS_ONLINE_STORE
            for online_store in online_stores
        ):
            _logger.info(
                "Not eligible for pgsql connection because feature tables are not published to Databricks Online Store."
            )
            return False
        if (
            len(set(online_store.extra_configs.host for online_store in online_stores))
            > 1
        ):
            _logger.info(
                "Not eligible for pgsql connection because feature tables are published to different online stores."
            )
            return False
        return True

    def _validate_input(
        self,
        df: pd.DataFrame,
        fully_overridden_feature_output_names: List[str],
        partially_overridden_feature_output_names: List[str],
    ):
        """
        Validates:
            - df contains exactly one column per SourceDataColumnInfo
            - df contains exactly one column per lookup key and per function input that are not
                provided by intermediate lookups or functions
            - df has no NaN lookup keys
        """
        req_source_column_names = [
            col_info.name for col_info in self.feature_spec.source_data_column_infos
        ]
        lookup_key_columns = set(
            itertools.chain.from_iterable(
                [
                    col_info.lookup_key
                    for col_info in self.feature_spec.feature_column_infos
                ]
            )
        )
        function_input_binding_columns = set(
            itertools.chain.from_iterable(
                [
                    col_info.input_bindings.values()
                    for col_info in self.feature_spec.on_demand_column_infos
                ]
            )
        )

        lookup_output_columns = set(
            [ci.output_name for ci in self.feature_spec.feature_column_infos]
        )
        function_output_columns = set(
            [ci.output_name for ci in self.feature_spec.on_demand_column_infos]
        )

        # All input and output excluding source columns.
        all_input_columns = function_input_binding_columns.union(lookup_key_columns)
        all_output_columns = lookup_output_columns.union(function_output_columns)

        all_req_columns = set(req_source_column_names).union(
            all_input_columns.difference(all_output_columns)
        )
        df_column_set = set(df.columns)
        missing_columns = [c for c in all_req_columns if c not in df_column_set]

        if missing_columns:
            raise ValueError(
                f"Input is missing columns {list(missing_columns)}. "
                f"\n\tThe following columns are required: {list(all_req_columns)}."
            )

        df_column_name_counts = Counter(df.columns)
        no_dup_columns_allowed_set = set(
            list(lookup_key_columns)
            + req_source_column_names
            + fully_overridden_feature_output_names
            + partially_overridden_feature_output_names
        )
        dup_columns = [
            col
            for col, count in df_column_name_counts.items()
            if count > 1 and col in no_dup_columns_allowed_set
        ]

        if dup_columns:
            raise ValueError(
                f"Input has duplicate columns: {dup_columns}"
                f"\n\tThe following column names must be unique: {no_dup_columns_allowed_set}"
            )

        # The set of columns from the input that are required as lookup keys.
        # This set doesn't include lookup keys that can be provided by intermediate lookup or
        # function result.
        lookup_key_columns_from_input = lookup_key_columns.difference(
            all_output_columns
        )
        lookup_key_df = df[list(lookup_key_columns_from_input)]
        cols_with_nulls = lookup_key_df.columns[lookup_key_df.isnull().any()].tolist()
        if cols_with_nulls:
            raise ValueError(
                f"Failed to lookup feature values due to null values for lookup_key columns "
                f"{cols_with_nulls}. The following columns cannot contain null values: "
                f"{lookup_key_columns}"
            )

    @timeout(seconds=TOTAL_REQUEST_TIMEOUT_SECONDS)
    def _augment_with_materialized_features(
        self,
        df: pd.DataFrame,
        features_to_lookup: List[FeatureColumnInfo],
        partially_overridden_feature_output_names: List[str],
        *,
        metrics: LookupClientMetrics = None,
    ) -> pd.DataFrame:
        """
        Adds materialized features to `df`. This function does not filter out ColumnInfos with
        include=False, however it will drop any columns in `df` that are not in the FeatureSpec.
        features_to_lookup MUST NOT include any fully overridden features already in df; however,
        partially overridden features are allowed and non-overridden cells will be looked up.

        For example, a FeatureSpec may include

            - revenue:
                include: false
                source: training_data
            - cost:
                include: false
                source: training_data
            - profit:
                udf_name: prod.math.subtract
                input_bindings:
                  x: sales
                  y: cost
                source: on_demand_feature

         In this case, the DataFrame returned by this method will include columns `revenue` and
         `cost`. This is required to compute the Feature Function `profit` at a later stage.

        :param df: Pandas DataFrame provided by user as model input. This is expected to contain
        columns for each SourceColumnInfo, and for each lookup key of a FeatureColumnInfo. Columns
        with the same name as FeatureColumnInfo output_names will override those features, meaning
        they will not be queried from the online store.
        :param features_to_lookup: List of FeatureColumnInfos to lookup from the online store.
        :param partially_overridden_feature_output_names: List of feature output names that are
        partially overridden in `df`. These features will be looked up from the online store, and
        only the cells that are not overridden will be looked up.
        :return: Pandas DataFrame containing all materialized features specified in the FeatureSpec,
          including features with include=False (see example above).
        """
        # Contains lookup result columns. Does not include fully overridden columns.
        feature_dfs = []
        partially_overridden_features_count = 0
        partially_overridden_feature_output_names_set = set(
            partially_overridden_feature_output_names
        )

        pk_lists = defaultdict(lambda: defaultdict(list))
        feature_column_infos_to_lookup_dict = defaultdict(lambda: defaultdict(list))
        lookup_clients = {}

        table_names_to_lookup = {info.table_name for info in features_to_lookup}
        all_lookup_columns = {info.output_name for info in features_to_lookup}

        # Query online store(s) for feature values
        ft_meta: FeatureTableMetadata
        for ft_name, ft_meta in self.ft_metadata.items():
            if ft_name not in table_names_to_lookup:
                continue
            if not self.is_model_eligible_for_batch_lookup:
                lookup_clients[
                    ft_meta.online_ft.online_feature_table_name
                ] = self.ft_to_lookup_client[ft_name]

            # Iterate through the lookup_keys for this feature table, each of which is used to
            # lookup a list of features
            lookup_key: LookupKeyType
            feature_column_infos: List[FeatureColumnInfo]
            for (
                lookup_key,
                feature_column_infos,
            ) in ft_meta.feature_col_infos_by_lookup_key.items():
                # Only lookup features that were specified in features_to_lookup
                feature_column_infos_to_lookup = [
                    fci
                    for fci in feature_column_infos
                    if fci.output_name in all_lookup_columns
                ]
                if len(feature_column_infos_to_lookup) == 0:
                    # All features were overridden in the input DataFrame
                    continue

                pk_lists[ft_meta.online_ft.online_feature_table_name][
                    lookup_key
                ] = get_primary_key_list(lookup_key, ft_meta.online_ft.primary_keys, df)
                feature_column_infos_to_lookup_dict[
                    ft_meta.online_ft.online_feature_table_name
                ][lookup_key] = feature_column_infos_to_lookup
        self._latency_debugger.add_latency("lk_prepare")

        # A nested defaultdict using table_name as first layer key and lookup_key as second layer
        # key. The actual inner value of the dict is of type pd.DataFrame. The structure looks like:
        # table_name: {
        #   lookup_key: DataFrame
        # }
        # TODO[ML-34263]: consider updating the default value type to match the actual usage.
        feature_values_dfs = defaultdict(lambda: defaultdict(list))
        if self.is_model_eligible_for_batch_lookup and pk_lists:
            feature_values_dfs = self._batch_lookup_and_rename_features(
                self.batch_lookup_client,
                pk_lists,
                feature_column_infos_to_lookup_dict,
                metrics=metrics,
            )
            self._latency_debugger.add_latency("lk_batch")
        else:
            for oft_name, pk_list_by_lookup_key in pk_lists.items():
                for lookup_key in pk_list_by_lookup_key.keys():
                    feature_values_dfs[oft_name][
                        lookup_key
                    ] = self._lookup_and_rename_features(
                        lookup_clients[oft_name],
                        pk_lists[oft_name][lookup_key],
                        feature_column_infos_to_lookup_dict[oft_name][lookup_key],
                        metrics=metrics,
                        feature_data_type_mapping=self.table_feature_data_type_mapping[
                            oft_name
                        ],
                    )
                    self._latency_debugger.add_latency("lk_table")

        for oft_name, feature_values_df_by_lookup_key in feature_values_dfs.items():
            for (
                lookup_key,
                feature_values_df,
            ) in feature_values_df_by_lookup_key.items():
                if feature_values_df.shape[0] != df.shape[0]:
                    raise Exception(
                        f"Internal Error: Expected {df.shape[0]} rows to be looked up from feature "
                        f"table {ft_name}, but found {feature_values_df.shape[0]}"
                    )

                # If any features were partially overridden, use the override values.
                # Filter all partially overridden feature output names down to those that
                # are in the feature table currently being processed.
                partially_overridden_feats = [
                    c
                    for c in feature_values_df.columns
                    if c in partially_overridden_feature_output_names_set
                ]
                if partially_overridden_feats:
                    # add partially overriden features to override count, count() returns the number
                    # of non-NaN values per column those are the values that will get overriden
                    partially_overridden_features_count += (
                        df[partially_overridden_feats].count().sum()
                    )
                    # For each cell of overridden column, use the overridden value if provided, else
                    # the value looked up from the online store.
                    partially_overridden_feats_df = df[
                        partially_overridden_feats
                    ].combine_first(feature_values_df[partially_overridden_feats])
                    feature_values_df[
                        partially_overridden_feats
                    ] = partially_overridden_feats_df[partially_overridden_feats]

                feature_dfs.append(feature_values_df)

        feature_names_looked_up = [feat for df in feature_dfs for feat in df.columns]
        used_partial_feature_names = (
            partially_overridden_feature_output_names_set.intersection(
                set(feature_names_looked_up)
            )
        )

        # Concatenate the following DataFrames, where N is the number of rows in `df`.
        #  1. feature_dfs - List of DataFrames, one per feature table looked up from the online
        #       store, containing partially overridden columns. Each DataFrame has N rows.
        #  2. non_overridden_columns_df - columns from `df` that are not partially overriden or are
        #       not looked-up in this iteration. DataFrame with N rows.
        non_overridden_columns_df = df.drop(columns=list(used_partial_feature_names))
        augmented_features = pd.concat(
            feature_dfs + [non_overridden_columns_df],
            axis=1,
        )
        self._latency_debugger.add_latency("lk_partial")

        if metrics:
            metrics.increase_metric(
                OVERRIDEN_FEATURE_COUNT, partially_overridden_features_count
            )

        return augmented_features

    def _compute_feature_functions(
        self,
        ff_inputs: pd.DataFrame,
        functions_to_apply: List[OnDemandColumnInfo],
        ff_overrides: Optional[pd.DataFrame],
    ) -> pd.DataFrame:
        """
        Executes Feature Functions specified by FeatureSpec.

        :param ff_inputs: A DataFrame containing all inputs required to evaluate Feature Functions.
          Additional columns are permitted.
        :param ff_overrides: A DataFrame containing overrides to Feature Functions. When a Feature
          Function output name is included as a column in `ff_overrides`, the function is computed
          only for rows where the override is `np.nan`.
        :return: DataFrame containing computed Feature Function columns
        """
        # TODO(ML-30619): Short-circuit Feature Function execution when overridden
        feature_function_df = self.feature_function_executor.execute_feature_functions(
            ff_inputs,
            functions_to_apply,
        )
        if ff_overrides is None:
            return feature_function_df

        ff_overrides_columns_set = set(ff_overrides.columns)
        overridden_ff_outputs = [
            odci.output_name
            for odci in functions_to_apply
            if odci.output_name in ff_overrides_columns_set
        ]
        overridden_ff_df = ff_overrides[overridden_ff_outputs]
        # As with materialized features, Feature Functions may be fully or partially
        # overridden, where a partial override means that some rows have null override value.
        # Merge the overridden values with the computed feature values:
        return overridden_ff_df.combine_first(feature_function_df)

    def _lookup_and_rename_features(
        self,
        lookup_client: OnlineLookupClient,
        primary_key_list: LookupKeyList,
        feature_column_infos: List[FeatureColumnInfo],
        *,
        metrics: LookupClientMetrics = None,
        feature_data_type_mapping: dict = None,
    ) -> pd.DataFrame:
        """
        Looks up features from a single feature table, then renames them. Feature metadata is
         specified via `feature_column_infos`.
        """
        feature_names = [fci.feature_name for fci in feature_column_infos]
        feature_values = lookup_client.lookup_features(
            primary_key_list, feature_names, metrics=metrics
        )
        feature_name_to_output_name = {
            fci.feature_name: fci.output_name for fci in feature_column_infos
        }

        feature_name_to_default_values = {}
        for fci in feature_column_infos:
            if fci.feature_name not in feature_data_type_mapping:
                _logger.warning(
                    f"_lookup_and_rename_features: Data type not found for unexpected feature {fci.feature_name}, default value will be ignored."
                )
                continue

            default_value = deserialize_default_value_to_data_type(
                fci.default_value_str, feature_data_type_mapping[fci.feature_name]
            )
            if default_value is not None:
                feature_name_to_default_values[fci.feature_name] = default_value

        if feature_name_to_default_values:
            feature_values = feature_values.fillna(value=feature_name_to_default_values)

        return feature_values.rename(feature_name_to_output_name, axis=1)

    def _batch_lookup_and_rename_features(
        self,
        batch_lookup_client: OnlineLookupClient,
        primary_key_lists: Dict[str, Dict[LookupKeyType, LookupKeyList]],
        feature_column_infos_dict: Dict[
            str, Dict[LookupKeyType, List[FeatureColumnInfo]]
        ],
        *,
        metrics: LookupClientMetrics = None,
    ) -> Dict[str, Dict[LookupKeyType, pd.DataFrame]]:
        """
        Looks up features from all the feature tables in batch, then renames them. Feature metadata
         is specified via `feature_column_infos`.
        """
        feature_names = defaultdict(lambda: defaultdict(list))
        feature_name_to_output_names = defaultdict(lambda: defaultdict(str))
        feature_name_to_default_values = defaultdict(lambda: defaultdict(lambda: None))
        for (
            oft_name,
            feature_column_infos_by_lookup_key,
        ) in feature_column_infos_dict.items():
            for (
                lookup_key,
                feature_column_infos,
            ) in feature_column_infos_by_lookup_key.items():
                feature_names[oft_name][lookup_key] = [
                    fci.feature_name for fci in feature_column_infos
                ]
                feature_name_to_output_names[oft_name][lookup_key] = {
                    fci.feature_name: fci.output_name for fci in feature_column_infos
                }

                feature_name_to_default_values[oft_name][lookup_key] = {}
                for fci in feature_column_infos:
                    if (
                        fci.feature_name
                        not in self.table_feature_data_type_mapping[oft_name]
                    ):
                        _logger.warning(
                            f"_batch_lookup_and_rename_features: Data type not found for unexpected feature {fci.feature_name}, default value will be ignored."
                        )
                        continue

                    default_value = deserialize_default_value_to_data_type(
                        fci.default_value_str,
                        self.table_feature_data_type_mapping[oft_name][
                            fci.feature_name
                        ],
                    )
                    if default_value is not None:
                        feature_name_to_default_values[oft_name][lookup_key][
                            fci.feature_name
                        ] = default_value

        feature_values = batch_lookup_client.batch_lookup_features(
            primary_key_lists, feature_names, metrics=metrics
        )

        for oft_name, feature_values_by_lookup_key in feature_values.items():
            for lookup_key, _ in feature_values_by_lookup_key.items():
                feature_values_by_oft_and_lookup_key = feature_values[oft_name][
                    lookup_key
                ]
                if feature_name_to_default_values[oft_name][lookup_key]:
                    feature_values_by_oft_and_lookup_key = (
                        feature_values_by_oft_and_lookup_key.fillna(
                            value=feature_name_to_default_values[oft_name][lookup_key]
                        )
                    )
                feature_values[oft_name][
                    lookup_key
                ] = feature_values_by_oft_and_lookup_key.rename(
                    feature_name_to_output_names[oft_name][lookup_key], axis=1
                )
        return feature_values

    def _resolve_online_stores(
        self,
        feature_tables: List[str],
        feature_tables_for_serving: FeatureTablesForServing,
    ) -> Dict[str, OnlineFeatureTable]:
        """
        :return: feature table name -> OnlineFeatureTable
        """
        all_fts_to_online_ft = {
            online_ft.feature_table_name: online_ft
            for online_ft in feature_tables_for_serving.online_feature_tables
        }

        missing = [ft for ft in feature_tables if ft not in all_fts_to_online_ft]
        if missing:
            raise Exception(
                f"Internal error: Online feature table information could not be found "
                f"for feature tables {missing}."
            )

        return {ft: all_fts_to_online_ft[ft] for ft in feature_tables}


def _load_pyfunc(path):
    """
    Called by ``pyfunc.load_pyfunc``.
    """
    return _FeatureStoreModelWrapper(path)



from typing import Any

import numpy as np

from databricks.ml_features_common.entities.data_type import DataType

# TODO: Add support for timestamp and date types
# - Decimal: currently not supported because model serving does not support it
# - Date, Timestamp: currently not supported because of PyYAML's "smart" parsing of dates and timestamp strings that breaks protobuf's ParseDict


def deserialize_default_value_to_data_type(
    value_string: str, data_type: DataType
) -> Any:
    """
    Deserialize a default value string representation to the specified data type.
    :param value_string: The string representation of the default value.
    :param data_type: The data type to which the value should be deserialized.
    :return: The deserialized value in the specified data type.
    """
    if not value_string or not data_type:
        return None

    if data_type == DataType.INTEGER:
        return np.int32(value_string)
    elif data_type == DataType.LONG:
        return np.int64(value_string)
    elif data_type == DataType.SHORT:
        return np.int16(value_string)
    elif data_type == DataType.FLOAT:
        return np.float32(value_string)
    elif data_type == DataType.DOUBLE:
        return np.float64(value_string)
    elif data_type == DataType.BOOLEAN:
        return value_string.lower() == "true"
    elif data_type == DataType.STRING:
        return value_string
    else:
        raise ValueError(f"Unsupported data type: {data_type} for default value")


def deserialize_default_value_primitive_type(
    value_string: str, data_type: DataType
) -> Any:
    """
    Deserialize a default value string representation to the specified Python primitive data types.
    :param value_string: The string representation of the default value.
    :param data_type: The data type to which the value should be deserialized.
    :return: The deserialized value in the specified Python data type.
    """
    if not value_string or not data_type:
        return None

    if (
        data_type == DataType.INTEGER
        or data_type == DataType.LONG
        or data_type == DataType.SHORT
    ):
        return int(value_string)
    elif data_type == DataType.FLOAT or data_type == DataType.DOUBLE:
        return float(value_string)
    elif data_type == DataType.BOOLEAN:
        return value_string.lower() == "true"
    elif data_type == DataType.STRING:
        return value_string
    else:
        raise ValueError(f"Unsupported data type: {data_type} for default value")


def serialize_default_value(value: Any) -> str:
    """
    Serialize a default value to its string representation.
    :param value: The default value to serialize.
    :return: The string representation of the default value.
    """
    if value is None:
        return ""
    elif isinstance(value, str):
        return value
    elif isinstance(
        value,
        (
            int,
            float,
            bool,
            np.int16,
            np.int32,
            np.int64,
            np.float32,
            np.float64,
        ),
    ):
        return str(value)
    else:
        raise ValueError(
            f"Unsupported type for default value serialization: {type(value)}"
        )



"""
Constants for brickstore integration.
"""
# Token should be updated every 30 mins by serving scheduler.
# Implement more frequent refresh for safety.
TOKEN_REFRESH_DURATION_SECONDS = 5 * 60
TOKEN_REFRESH_JITTER_MAX_SECONDS = 10
TOKEN_REFRESH_BASE_RETRY_DELAY_SECONDS = 2
# Max retry attempts == 9, meaning at most 10 attempts with total retry time 1024 seconds (17 minutes).
TOKEN_REFRESH_MAX_RETRIES = 9

# Mount folder for served model secrets.
SECRET_MOUNT_LOCATION = "/var/credentials-secret"

# HTTP gateway oauth token
BRICKSTORE_OAUTH_TOKEN_FILE_NAME = "brickstore-feature-lookup"
BRICKSTORE_OAUTH_TOKEN_FILE_PATH = (
    SECRET_MOUNT_LOCATION + "/" + BRICKSTORE_OAUTH_TOKEN_FILE_NAME
)

# PgSQL oauth token
LAKEBASE_OAUTH_TOKEN_FILE_NAME = "lakebase-sql-feature-lookup"
LAKEBASE_OAUTH_TOKEN_FILE_PATH = (
    SECRET_MOUNT_LOCATION + "/" + LAKEBASE_OAUTH_TOKEN_FILE_NAME
)



import functools
import logging
import os
import signal

TOTAL_REQUEST_TIMEOUT_SECONDS_ENV_VAR = "FEATURE_STORE_TOTAL_REQUEST_TIMEOUT_SECONDS"
# The default total request timeout in lookup client is 90% of the default model serving request timeout
DEFAULT_TOTAL_REQUEST_TIMEOUT_SECONDS = int(0.9 * 120)

_logger = logging.getLogger(__name__)


def get_total_request_timeout_seconds():
    model_request_timeout = os.environ.get(TOTAL_REQUEST_TIMEOUT_SECONDS_ENV_VAR)
    if model_request_timeout:
        try:
            return int(model_request_timeout)
        except ValueError:
            _logger.warning(
                f"Unable to cast {model_request_timeout} to int. Using default timeout."
            )
            return DEFAULT_TOTAL_REQUEST_TIMEOUT_SECONDS
    else:
        return DEFAULT_TOTAL_REQUEST_TIMEOUT_SECONDS


TOTAL_REQUEST_TIMEOUT_SECONDS = get_total_request_timeout_seconds()


def timeout(seconds):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise Exception(
                f"Exceeded maximum wait time of {seconds} seconds to lookup the features"
            )

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # store the current signal handler
            original_handler = signal.getsignal(signal.SIGALRM)
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                return func(*args, **kwargs)
            finally:
                signal.alarm(0)
                # send signal to the original handler
                signal.signal(signal.SIGALRM, original_handler)

        return wrapper

    return decorator



import os
from functools import wraps

import mlflow
from mlflow.entities import SpanType
from packaging import version

ENABLE_FEATURE_TRACING = "ENABLE_FEATURE_TRACING"


def is_feature_tracing_enabled() -> bool:
    # environment variables stored as string, rather than boolean
    # MLflow tracing is only available after Mlflow 2.14.0
    current_mlflow_version = version.parse(mlflow.__version__)
    minimum_mlflow_version = version.parse("2.14.0")
    return (
        os.getenv(ENABLE_FEATURE_TRACING) != "false"
        and current_mlflow_version >= minimum_mlflow_version
    )


def trace_latency(span_name: str):
    """
    A decorator for simple tracing. Any method decorated by this decorator
    will add a span with the given name. Compared with the official mlflow.trace,
    this version does not record method inputs or outputs.
    """

    def decorator(func):
        if is_feature_tracing_enabled():
            # Apply mlflow.trace directly if tracing is enabled
            return mlflow.trace(name=span_name, span_type=SpanType.RETRIEVER)(func)
        else:
            # Otherwise, return the original function unchanged
            return func

    return decorator



import os

from databricks.feature_store.utils.lookup_client_envvars import (
    FEATURE_SERVING_CONNECTION_POOL_SIZE,
)

# Setting a max connection pool size to avoid exhasusting all the connections on Lakebase.
# quickly. Each lakebase replica has a limit of 1000 connections. The maximum number of
# model serving concurrency supported is:
# (1000 * {number of Lakebase read replicas}) / ({connection pool size} * {number of logical databases})
# If provisioning concurrency higher than this, new pods cannot not start because of the connection limit.
DEFAULT_MAX_CONNECTION_POOL_SIZE = 10
# The minimum connection pool size 3 is tested with single table lookup use cases with
# for high throughput use cases for best availability.
DEFAULT_MIN_CONNECTION_POOL_SIZE = 3


def get_lakebase_connection_pool_size(number_of_tables: int) -> int:
    """
    Get the connection pool size for Lakebase PgSql.

    :param number_of_tables: The number of tables to lookup. This should be the
        number of tables that are looked up in parallel. In case of a DAG execution,
        this should be the biggest number of tables in one execution group.
    """
    if number_of_tables is None:
        raise Exception("number_of_tables is required for Lakebase PgSql connections.")
    if FEATURE_SERVING_CONNECTION_POOL_SIZE in os.environ:
        return int(os.environ[FEATURE_SERVING_CONNECTION_POOL_SIZE])
    else:
        return max(
            DEFAULT_MIN_CONNECTION_POOL_SIZE,
            min(DEFAULT_MAX_CONNECTION_POOL_SIZE, number_of_tables),
        )



import json
import logging
import os
import random
import threading
import time

from databricks.feature_store.utils.lakebase_constants import (
    TOKEN_REFRESH_BASE_RETRY_DELAY_SECONDS,
    TOKEN_REFRESH_DURATION_SECONDS,
    TOKEN_REFRESH_JITTER_MAX_SECONDS,
    TOKEN_REFRESH_MAX_RETRIES,
)
from databricks.feature_store.utils.logging_utils import get_logger

_logger = get_logger(__name__, log_level=logging.INFO)


class OAuthTokenManager:
    def __init__(self, oauth_token_file_path: str, password_override: str = None):
        self._oauth_token = None
        self._password = password_override
        self._oauth_token_file_path = oauth_token_file_path
        self.refresh_oauth_token()

        self._run_token_refresh_thread = True
        self._token_refresh_thread = threading.Thread(
            target=self._refresh_oauth_token_loop
        )
        # Enable daemon thread to ensure the thread exits when the main thread exits.
        self._token_refresh_thread.daemon = True

    def start_token_refresh_thread(self):
        self._token_refresh_thread.start()

    def refresh_oauth_token(self, retry_count=0):
        """
        Periodically refresh from mounted secret file. Upstream validation ensures this path
        should exist.
        """
        if self._password:
            # If password is provided by endpoint deployer and is NOT expected to be
            # rotated, use it instead of oauth token.
            return
        should_retry = False

        if not os.path.exists(self._oauth_token_file_path):
            # File doesn't exist yet, trigger retry logic
            if retry_count >= TOKEN_REFRESH_MAX_RETRIES:
                raise Exception("OAuth token file not found and max retries exceeded.")
            should_retry = True
        else:
            with open(self._oauth_token_file_path, "r") as f:
                try:
                    oauth_dict = json.load(f)
                    ret_val = oauth_dict["OAUTH_TOKEN"][0]["oauthTokenValue"]
                    self._oauth_token = ret_val
                except Exception:
                    # Remediation for potential race condition in which the read occurs
                    # simultaneously with the secret mount update, resulting in a malformed
                    # token
                    if retry_count >= TOKEN_REFRESH_MAX_RETRIES:
                        raise Exception(
                            "Invalid online store credential configuration."
                        )
                    should_retry = True
        if should_retry:
            # Exponential backoff with jitter
            jitter = random.randint(0, TOKEN_REFRESH_JITTER_MAX_SECONDS)
            time.sleep(
                TOKEN_REFRESH_BASE_RETRY_DELAY_SECONDS * (2**retry_count) + jitter
            )
            if retry_count > 3:
                _logger.warning(
                    f"Retrying to read oauth token for the {retry_count - 1} time."
                )
            self.refresh_oauth_token(retry_count + 1)

    def _refresh_oauth_token_loop(self):
        """
        Periodically refresh from mounted secret file. Upstream validation ensures this path
        should exist.
        """
        while self._run_token_refresh_thread:
            # Adding a jitter to avoid all threads refreshing at the same time.
            jitter = random.randint(0, TOKEN_REFRESH_JITTER_MAX_SECONDS)
            time.sleep(TOKEN_REFRESH_DURATION_SECONDS + jitter)
            self.refresh_oauth_token()

    def get_oauth_token_or_password(self):
        if self._password:
            # Password is provided to override the oauth token.
            return self._password
        if not self._oauth_token:
            self.refresh_oauth_token()
        return self._oauth_token



import os
from typing import Any, Dict, List, Type, Union

import mlflow
from google.protobuf.json_format import MessageToDict, ParseDict
from mlflow.utils.file_utils import TempDir
from typing_extensions import override

from databricks.ml_features_common.entities._feature_spec_base import _FeatureSpecBase
from databricks.ml_features_common.entities.column_info import ColumnInfo
from databricks.ml_features_common.entities.data_type import DataType
from databricks.ml_features_common.entities.feature_column_info import FeatureColumnInfo
from databricks.ml_features_common.entities.feature_spec_constants import (
    BOUND_TO,
    DATA_TYPE,
    FEATURE_COLUMN_INFO,
    FEATURE_STORE,
    INCLUDE,
    INPUT_BINDINGS,
    INPUT_COLUMNS,
    INPUT_FUNCTIONS,
    INPUT_TABLES,
    NAME,
    ON_DEMAND_COLUMN_INFO,
    ON_DEMAND_FEATURE,
    OUTPUT_NAME,
    PARAMETER,
    SERIALIZATION_VERSION,
    SOURCE,
    SOURCE_DATA_COLUMN_INFO,
    TABLE_NAME,
    TOPOLOGICAL_ORDERING,
    TRAINING_DATA,
    UDF_NAME,
)
from databricks.ml_features_common.entities.feature_table_info import FeatureTableInfo
from databricks.ml_features_common.entities.feature_tables_for_serving import (
    FeatureTablesForServing,
)
from databricks.ml_features_common.entities.function_info import FunctionInfo
from databricks.ml_features_common.entities.on_demand_column_info import (
    OnDemandColumnInfo,
)
from databricks.ml_features_common.entities.source_data_column_info import (
    SourceDataColumnInfo,
)
from databricks.ml_features_common.protos.feature_spec_pb2 import (
    FeatureSpec as ProtoFeatureSpec,
)
from databricks.ml_features_common.utils import utils_common
from databricks.ml_features_common.utils.utils_common import is_artifact_uri
from databricks.ml_features_common.utils.yaml_utils import read_yaml, write_yaml

# Change log for serialization version. Please update for each serialization version.
# 1. Initial.
# 2. (2021/06/16): Record feature_store_client_version to help us make backward compatible changes in the future.
# 3. (2021/08/25): Record table_id to handle feature table lineage stability if tables are deleted.
# 4. (2021/09/25): Record timestamp_lookup_key to handle point-in-time lookups.
# 5. (2021/02/15): Record include flag for column info if False.
#                  Record input functions as FunctionInfo and function computation as OnDemandColumnInfo.
#                  Remove redundant fields: table_name from table_infos, output_name from column_infos.
# 6. (2023/04/21): Record lookback_window in table info for point-in-time lookups.
# 7. (2023/05/05): Record the Spark data type for all columns to track model signatures.
# 8. (2023/08/14): Record the topological_ordering for all columns to support chained transform and lookup.
# 9. (2023/09/11): Change the type of lookback_window from int to double for sub-second values


class FeatureSpec(_FeatureSpecBase):
    def __init__(
        self,
        column_infos: List[ColumnInfo],
        table_infos: List[FeatureTableInfo],
        function_infos: List[FunctionInfo],
        workspace_id: int,
        feature_store_client_version: str,
        serialization_version: int,
    ):
        super().__init__(
            serialization_version, workspace_id, feature_store_client_version
        )
        self._column_infos = column_infos
        self._table_infos = table_infos
        self._function_infos = function_infos

        # Perform validations
        self._validate_column_infos()
        self._validate_table_infos()
        self._validate_function_infos()

    def _validate_column_infos(self):
        if not self.column_infos:
            raise ValueError("column_infos must be non-empty.")

        for column_info in self.column_infos:
            if not isinstance(column_info, ColumnInfo):
                raise ValueError(
                    f"Expected all elements of column_infos to be instances of ColumnInfo. "
                    f"'{column_info}' is of the wrong type."
                )
            if (
                self._serialization_version >= 8
                and column_info.topological_ordering is not None
            ):
                ordering = column_info.topological_ordering
                if not isinstance(ordering, int) or ordering < 0:
                    raise ValueError(
                        "The topological_ordering of column_info must be non non-negative integers."
                    )

    def _validate_table_infos(self):
        if self.table_infos is None:
            raise ValueError("Internal Error: table_infos must be provided.")

        # table_infos should not be duplicated
        utils_common.validate_strings_unique(
            [table_info.table_name for table_info in self.table_infos],
            "Internal Error: Expect all table_names in table_infos to be unique. Found duplicates {}",
        )

        # Starting FeatureSpec v3, unique table names in table_infos must match those in column_infos.
        if self.serialization_version >= 3:
            unique_table_names = set(
                [table_info.table_name for table_info in self.table_infos]
            )
            unique_column_table_names = set(
                [fci.table_name for fci in self.feature_column_infos]
            )
            if unique_table_names != unique_column_table_names:
                raise Exception(
                    f"Internal Error: table_names from table_infos {sorted(unique_table_names)} "
                    f"must match those from column_infos {sorted(unique_column_table_names)}"
                )

    def _validate_function_infos(self):
        if self.function_infos is None:
            raise ValueError("Internal Error: function_infos must be provided.")

        # function_infos should not be duplicated
        utils_common.validate_strings_unique(
            [function_info.udf_name for function_info in self.function_infos],
            "Internal Error: Expect all udf_names in function_infos to be unique. Found duplicates {}",
        )

        # Unique UDF names in function_infos must match those in column_infos.
        # No version check is required as both fields were added simultaneously in FeatureSpec v5.
        unique_udf_names = set(
            [function_info.udf_name for function_info in self.function_infos]
        )
        unique_column_udf_names = set(
            [odci.udf_name for odci in self.on_demand_column_infos]
        )
        if unique_udf_names != unique_column_udf_names:
            raise Exception(
                f"Internal Error: udf_names from function_infos {sorted(unique_udf_names)} "
                f"must match those from column_infos {sorted(unique_column_udf_names)}"
            )

    @property
    def column_infos(self):
        return self._column_infos

    @property
    def table_infos(self):
        return self._table_infos

    @property
    def function_infos(self):
        return self._function_infos

    def _get_infos_of_type(
        self,
        info_type: Union[
            Type[SourceDataColumnInfo],
            Type[FeatureColumnInfo],
            Type[OnDemandColumnInfo],
        ],
    ):
        """
        Helper method to return the ColumnInfo.info subinfo field based on its type.
        """
        return [
            column_info.info
            for column_info in self.column_infos
            if isinstance(column_info.info, info_type)
        ]

    @property
    def source_data_column_infos(self) -> List[SourceDataColumnInfo]:
        return self._get_infos_of_type(SourceDataColumnInfo)

    @property
    def feature_column_infos(self) -> List[FeatureColumnInfo]:
        return self._get_infos_of_type(FeatureColumnInfo)

    @property
    def on_demand_column_infos(self) -> List[OnDemandColumnInfo]:
        return self._get_infos_of_type(OnDemandColumnInfo)

    @classmethod
    def from_proto(cls, feature_spec_proto):
        # Serialization version is not deserialized from the proto as there is currently only one
        # possible version.
        column_infos = [
            ColumnInfo.from_proto(column_info_proto)
            for column_info_proto in feature_spec_proto.input_columns
        ]
        table_infos = [
            FeatureTableInfo.from_proto(table_info_proto)
            for table_info_proto in feature_spec_proto.input_tables
        ]
        function_infos = [
            FunctionInfo.from_proto(function_info_proto)
            for function_info_proto in feature_spec_proto.input_functions
        ]
        return cls(
            column_infos=column_infos,
            table_infos=table_infos,
            function_infos=function_infos,
            workspace_id=feature_spec_proto.workspace_id,
            feature_store_client_version=feature_spec_proto.feature_store_client_version,
            serialization_version=feature_spec_proto.serialization_version,
        )

    def has_feature_renaming(self):
        for fc in self.feature_column_infos:
            if fc.output_name != fc.feature_name:
                return True
        for oc in self.on_demand_column_infos:
            if oc.output_name != oc.udf_name:
                return True
        return False

    def has_default_value(self):
        for fc in self.feature_column_infos:
            if fc.default_value_str:
                return True
        return False

    def has_dag(self):
        """
        Check if the FeatureSpec has a DAG. If a DAG is detected the feature lookup needs
        to be performed in multiple steps.
        """
        source_column_names = set(info.name for info in self.source_data_column_infos)
        for column_info in self.column_infos:
            if isinstance(column_info.info, FeatureColumnInfo):
                for lookup_key in column_info.info.lookup_key:
                    if lookup_key not in source_column_names:
                        return True
            elif isinstance(column_info.info, OnDemandColumnInfo):
                for parameter in column_info.info.input_bindings.keys():
                    if parameter not in source_column_names:
                        return True
        return False

    def to_proto(self):
        proto_feature_spec = ProtoFeatureSpec()
        for column_info in self.column_infos:
            proto_feature_spec.input_columns.append(column_info.to_proto())
        for table_info in self.table_infos:
            proto_feature_spec.input_tables.append(table_info.to_proto())
        for function_info in self.function_infos:
            proto_feature_spec.input_functions.append(function_info.to_proto())
        proto_feature_spec.serialization_version = self.serialization_version
        proto_feature_spec.workspace_id = self.workspace_id
        proto_feature_spec.feature_store_client_version = (
            self._feature_store_client_version
        )
        return proto_feature_spec

    @staticmethod
    def _input_columns_proto_to_yaml_dict(column_info: Dict[str, Any]):
        """
        Converts a single ColumnInfo's proto dict to the expected element in FeatureSpec YAML's input_columns.
        To keep the YAML clean, unnecessary fields are removed (e.g. SourceDataColumnInfo.name field, ColumnInfo.include when True).

        Example of a column_info transformation. Note that "name" and "include" attributes were excluded.
        {"source_data_column_info": {"name": "source_column"}, "include": True} -> {"source_column": {"source": "training_data"}}

        Order of elements in the YAML dict should be:
        1. Attributes present in ColumnInfo.info, using the proto field order
        2. Remaining attributes of ColumnInfo, using the proto field order
        3. Feature Store source type
        """
        # Parse oneof field ColumnInfo.info level attributes as column_info_attributes; record column_name, source
        if SOURCE_DATA_COLUMN_INFO in column_info:
            column_info_attributes = column_info[SOURCE_DATA_COLUMN_INFO]
            # pop NAME attribute and use as the YAML key for this column_info to avoid redundancy in YAML
            column_name, source = column_info_attributes.pop(NAME), TRAINING_DATA
        elif FEATURE_COLUMN_INFO in column_info:
            column_info_attributes = column_info[FEATURE_COLUMN_INFO]
            # pop OUTPUT_NAME attribute and use as the YAML key for this column_info to avoid redundancy in YAML
            column_name, source = column_info_attributes.pop(OUTPUT_NAME), FEATURE_STORE
        elif ON_DEMAND_COLUMN_INFO in column_info:
            column_info_attributes = column_info[ON_DEMAND_COLUMN_INFO]
            # Map InputBindings message dictionary to {parameter: bound_to} KV dictionary if defined
            if INPUT_BINDINGS in column_info_attributes:
                column_info_attributes[INPUT_BINDINGS] = {
                    ib[PARAMETER]: ib[BOUND_TO]
                    for ib in column_info_attributes[INPUT_BINDINGS]
                }
            # pop OUTPUT_NAME attribute and use as the YAML key for this column_info to avoid redundancy in YAML
            column_name, source = (
                column_info_attributes.pop(OUTPUT_NAME),
                ON_DEMAND_FEATURE,
            )
        else:
            raise ValueError(
                f"Expected column_info to be keyed by a valid ColumnInfo.info type. "
                f"'{column_info}' has key '{list(column_info)[0]}'."
            )

        # Parse and insert ColumnInfo level attributes
        # Note: the ordering of fields in the result yaml file is undefined but in reality, they are
        # in the same order as they are added in the column_info_attributes dict.

        # DATA_TYPE is supported starting FeatureSpec v7 and is not guaranteed to exist.
        if DATA_TYPE in column_info:
            column_info_attributes[DATA_TYPE] = column_info[DATA_TYPE]
        if not column_info[INCLUDE]:
            column_info_attributes[INCLUDE] = False
        # TOPOLOGICAL_ORDERING is supported starting FeatureSpec v8.
        if TOPOLOGICAL_ORDERING in column_info:
            column_info_attributes[TOPOLOGICAL_ORDERING] = column_info[
                TOPOLOGICAL_ORDERING
            ]

        # Insert source; return YAML keyed by column_name
        column_info_attributes[SOURCE] = source
        return {column_name: column_info_attributes}

    @override
    def _to_dict(self):
        """
        Convert FeatureSpec to a writeable YAML artifact. Uses MessageToDict to convert FeatureSpec proto to dict.
        Sanitizes and modifies the dict as follows:
        1. Remove redundant or unnecessary information for cleanliness in the YAML
        2. Modifies the dict to be of the format {column_name: column_attributes_dict}

        :return: Sanitized FeatureSpec dictionary of {column_name: column_attributes}
        """
        yaml_dict = MessageToDict(self.to_proto(), preserving_proto_field_name=True)
        yaml_dict[INPUT_COLUMNS] = [
            self._input_columns_proto_to_yaml_dict(column_info)
            for column_info in yaml_dict[INPUT_COLUMNS]
        ]

        if INPUT_TABLES in yaml_dict:
            # pop TABLE_NAME attribute and use as the YAML key for each table_info to avoid redundancy in YAML
            yaml_dict[INPUT_TABLES] = [
                {table_info.pop(TABLE_NAME): table_info}
                for table_info in yaml_dict[INPUT_TABLES]
            ]
        if INPUT_FUNCTIONS in yaml_dict:
            # pop UDF_NAME attribute and use as the YAML key for each table_info to avoid redundancy in YAML
            yaml_dict[INPUT_FUNCTIONS] = [
                {function_info.pop(UDF_NAME): function_info}
                for function_info in yaml_dict[INPUT_FUNCTIONS]
            ]

        # For readability, place SERIALIZATION_VERSION last in the dictionary.
        yaml_dict[SERIALIZATION_VERSION] = yaml_dict.pop(SERIALIZATION_VERSION)
        return yaml_dict

    @staticmethod
    def _input_columns_yaml_to_proto_dict(column_info: Dict[str, Any]):
        """
        Convert the FeatureSpec YAML dictionary to the expected ColumnInfo proto dictionary.

        Example of a column_info transformation.
        {"source_column": {"source": "training_data"}} -> {"source_data_column_info": {"name": "source_column"}}
        """
        if len(column_info) != 1:
            raise ValueError(
                f"Expected column_info dictionary to only have one key, value pair. "
                f"'{column_info}' has length {len(column_info)}."
            )
        column_name, column_data = list(column_info.items())[0]
        if not column_data:
            raise ValueError(
                f"Expected values of '{column_name}' dictionary to be non-empty."
            )
        if SOURCE not in column_data:
            raise ValueError(
                f"Expected values of column_info dictionary to include the source. No source found "
                f"for '{column_name}'."
            )

        # Parse oneof field ColumnInfo.info level attributes
        source = column_data.pop(SOURCE)
        if source == TRAINING_DATA:
            column_data[NAME] = column_name
            column_info_dict = {SOURCE_DATA_COLUMN_INFO: column_data}
        elif source == FEATURE_STORE:
            column_data[OUTPUT_NAME] = column_name
            column_info_dict = {FEATURE_COLUMN_INFO: column_data}
        elif source == ON_DEMAND_FEATURE:
            column_data[OUTPUT_NAME] = column_name
            # Map {parameter_val: bound_to_val} dictionary to InputBindings(parameter, bound_to) message dictionary.
            column_data[INPUT_BINDINGS] = [
                {PARAMETER: parameter, BOUND_TO: bound_to}
                for parameter, bound_to in column_data.get(INPUT_BINDINGS, {}).items()
            ]
            column_info_dict = {ON_DEMAND_COLUMN_INFO: column_data}
        else:
            raise ValueError(
                f"Internal Error: Expected column_info to have source matching oneof ColumnInfo.info. "
                f"'{column_info}' has source of '{source}'."
            )

        # Parse ColumnInfo level attributes
        # TOPOLOGICAL_ORDERING is supported starting FeatureSpec v8.
        if TOPOLOGICAL_ORDERING in column_data:
            column_info_dict[TOPOLOGICAL_ORDERING] = column_data.pop(
                TOPOLOGICAL_ORDERING
            )
        # DATA_TYPE is supported starting FeatureSpec v7 and is not guaranteed to exist.
        if DATA_TYPE in column_data:
            column_info_dict[DATA_TYPE] = column_data.pop(DATA_TYPE)
        # INCLUDE is supported starting FeatureSpec v5 and only present in the YAML when INCLUDE = False
        if INCLUDE in column_data:
            column_info_dict[INCLUDE] = column_data.pop(INCLUDE)
        return column_info_dict

    @classmethod
    def _from_dict(cls, spec_dict):
        """
        Convert YAML artifact to FeatureSpec. Transforms YAML artifact to dict keyed by
        source_data_column_info or feature_column_info, such that ParseDict can convert the dict to
        a proto message, and from_proto can convert the proto message to a FeatureSpec object
        :return: :py:class:`~databricks.ml_features_common.entities.feature_spec.FeatureSpec`
        """
        if INPUT_COLUMNS not in spec_dict:
            raise ValueError(
                f"{INPUT_COLUMNS} must be a key in {cls.FEATURE_ARTIFACT_FILE}."
            )
        if not spec_dict[INPUT_COLUMNS]:
            raise ValueError(
                f"{INPUT_COLUMNS} in {cls.FEATURE_ARTIFACT_FILE} must be non-empty."
            )
        spec_dict[INPUT_COLUMNS] = [
            cls._input_columns_yaml_to_proto_dict(column_info)
            for column_info in spec_dict[INPUT_COLUMNS]
        ]

        # feature_spec.yaml doesn't include input_tables, input_functions if any are true:
        # 1. The YAML is written by an older client that does not support the functionality.
        # 2. The FeatureSpec does not contain FeatureLookups (input_tables), FeatureFunctions (input_functions).
        input_tables = []
        for input_table in spec_dict.get(INPUT_TABLES, []):
            table_name, attributes = list(input_table.items())[0]
            input_tables.append({TABLE_NAME: table_name, **attributes})
        spec_dict[INPUT_TABLES] = input_tables

        input_functions = []
        for input_function in spec_dict.get(INPUT_FUNCTIONS, []):
            udf_name, attributes = list(input_function.items())[0]
            input_functions.append({UDF_NAME: udf_name, **attributes})
        spec_dict[INPUT_FUNCTIONS] = input_functions

        return cls.from_proto(
            ParseDict(spec_dict, ProtoFeatureSpec(), ignore_unknown_fields=True)
        )

    @classmethod
    def _read_file(cls, path: str):
        """
        Read the YAML artifact from a file path.
        """
        parent_dir, file = os.path.split(path)
        spec_dict = read_yaml(parent_dir, file)
        return cls._from_dict(spec_dict)

    @classmethod
    def load(cls, path: str):
        """
        Deprecated: Use feature_spec_utils::load_feature_spec instead.

        Load the FeatureSpec YAML artifact in the provided root directory (at path/feature_spec.yaml).

        :param path: Root path to the YAML artifact. This can be a MLflow artifact path or file path.
        :return: :py:class:`~databricks.ml_features_common.entities.feature_spec.FeatureSpec`
        """
        # Create the full file path to the FeatureSpec.
        path = os.path.join(path, cls.FEATURE_ARTIFACT_FILE)

        if is_artifact_uri(path):
            with TempDir() as tmp_location:
                # Returns a file and not directory since the artifact_uri is a single file.
                local_path = mlflow.artifacts.download_artifacts(
                    artifact_uri=path, dst_path=tmp_location.path()
                )
                return FeatureSpec._read_file(local_path)
        else:
            return FeatureSpec._read_file(path)

    @classmethod
    def from_feature_tables_for_serving(
        cls,
        feature_tables_for_serving: FeatureTablesForServing,
    ) -> "FeatureSpec":
        """
        Constructs a FeatureSpec object from FeatureTablesForServing for declarative features.

        This method extracts feature metadata from online feature tables and constructs a
        FeatureSpec programmatically. It creates ColumnInfo objects for each feature in each
        online feature table, mapping primary keys and timestamp keys, and creates
        FeatureTableInfo objects for each unique table name.

        Notes:
        1. For declarative features, we use the online table name as the table name in the
           FeatureSpec. This is because declarative features might not be materialized in an
           offline table (e.g., for streaming cases), and the table name is only used to look
           up online tables.
        2. We do not support renaming for declarative features. Therefore,
           we assume the declarative feature name is unique and matches the column name in the
           online table. It also matches the column name in the dataframe resulting from the lookup.
           (i.e., feature_name == output_name).
        3. We do not support default values for declarative features.

        :param feature_tables_for_serving: FeatureTablesForServing object containing online
            feature table definitions with features, primary keys, and timestamp keys.
        :param workspace_id: Workspace ID for the FeatureSpec. Defaults to 0.
        :param feature_store_client_version: Feature Store client version. Defaults to empty string.
        :return: FeatureSpec object with column_infos, table_infos, and empty function_infos
            (since declarative features don't use feature functions).
        """

        column_infos = [
            ColumnInfo(
                info=FeatureColumnInfo(
                    table_name=online_feature_table.online_feature_table_name,
                    feature_name=feature.name,
                    lookup_key=[key.name for key in online_feature_table.primary_keys],
                    output_name=feature.name,
                    timestamp_lookup_key=[
                        key.name for key in online_feature_table.timestamp_keys
                    ],
                ),
                include=True,
                data_type=DataType.to_string(feature.data_type)
                if feature.data_type
                else None,
            )
            for online_feature_table in feature_tables_for_serving.online_feature_tables
            for feature in online_feature_table.features
        ]

        # Create FeatureTableInfo objects for each unique table name
        # FeatureSpec validation requires table_infos to match table names in column_infos
        unique_table_names = set(
            online_feature_table.online_feature_table_name
            for online_feature_table in feature_tables_for_serving.online_feature_tables
        )
        table_infos = [
            FeatureTableInfo(
                table_name=table_name,
                table_id=None,
            )
            for table_name in unique_table_names
        ]

        return cls(
            column_infos=column_infos,
            table_infos=table_infos,
            function_infos=[],
            workspace_id=None,
            feature_store_client_version=None,
            serialization_version=cls.SERIALIZATION_VERSION_NUMBER,
        )
